
// model:
// A view has
//            a set of 0 or more attached clients that are attached exclusively to it,
//            a set of nodes that belong exclusively to it,
//        and a set of edges that belong exclusively to it.
// A client has 
//              a user name,
//              0 or 1 views to which it is attached.
// A node has
//            0 or 1 attached clients,
//            a view to which it belongs,
//            a text string,
//            an x coordinate, where 0 <= x < 1
//        and a y coordinate, where 0 <= y < 1.
// An edge has
//             0 or 1 attached clients,
//             a view to which it belongs,
//             a subject node,
//             a predicate node,
//         and an object node.

// data representation layer:

// Each type of thing (view, client, node, or edge) has its own Object constructor and ID-to-object map Object.

// The objects have properties of type Watchable so that the view controlling layer can add callbacks to update the view as the model changes.

function Watchable ( value, defaultValue ) {
  
  if ( typeof value == 'undefined' ) {
    value = defaultValue;
  }
  
  var changeCallbacks = [];
  
  this.set = function ( newValue ) {
    // Since setting the value could involve a lot of activity,
    // only proceed if a change is actually occurring.
    if ( newValue === value ) {
      return;
    }
    var oldValue = value;
    value = newValue;
    function callChangeCallback ( callback ) {
      callback(oldValue,newValue);
    }// end of function callChangeCallback
    changeCallbacks.forEach(callChangeCallback);
  };// end of this.set
  
  this.get = function () {
    return value;
  };// end of this.get
  
  // Convenience methods for numeric values:
  
  this.add = function (number) {
    this.set( this.get() + number );
  };// end of this.add
  
  this.increment = function () {
    this.add(1);
  };// end of this.increment
  
  this.decrement = function () {
    this.add(-1);
  };// end of this.decrement
  
  this.addChangeCallback = function ( callback ) {
    changeCallbacks.push(callback);
  };// end of this.addChangeCallback
  
  // Note we do not check whether indexOf returns -1 in remove...,
  // because that would reflect a programming error withing the client-side script,
  // not bad data from the server.
  this.removeChangeCallback = function ( callback ) {
    changeCallbacks.splice( changeCallbacks.indexOf(callback), 1 );
  };// end of this.removeChangeCallback
  
}// end of function Watchable

function View (id, clientCount, nodeCount, edgeCount) {
  
  this.getId = function () {
    return id;
  };// end of this.getId
  
  this.clientCountWatch = new Watchable(clientCount,0);
  this.nodeCountWatch = new Watchable(nodeCount,0);
  this.edgeCountWatch = new Watchable(edgeCount,0);
  
  this.cleanup = function () {
    // Views require no cleanup, but we are giving WatchableMap something to call.
  };// end of this.cleanup
  
}// end of function View

function Client ( id, userName, view ) {
  
  this.getId = function () {
    return id;
  };// end of this.getId
  
  if ( view !== null ) {
    view.clientCountWatch.increment();
  }
  
  this.userNameWatch = new Watchable(userName,'anonymous');
  this.viewWatch = new Watchable(view,null);
  
  function updateClientCounts ( oldView, newView ) {
    if ( oldView !== null ) {
      oldView.clientCountWatch.decrement();
    }
    if ( newView !== null ) {
      newView.clientCountWatch.increment();
    }
  }// end of function updateClientCounts
  this.viewWatch.addChangeCallback(updateClientCounts);
  
  this.cleanup = function () {
    this.viewWatch.set(null);
  };// end of this.cleanup
  
}// end of function Client

var MIN_COORDINATE = 0;
var MAX_COORDINATE = 1;
var MIDDLE_COORDINATE = MIN_COORDINATE + (MAX_COORDINATE - MIN_COORDINATE)/2;
var DEFAULT_TEXT = "Click to select.";
var DEFAULT_URL = 'http://elvis.biomed.drexel.edu/post-text-map-preview/client/tables.html';

function Node ( id, text, url, x, y, view, client, subjectEdgeCount, predicateEdgeCount, objectEdgeCount ) {
  
  this.getId = function () {
    return id;
  };// end of this.getId
  
  if ( view !== null ) {
    view.nodeCountWatch.increment();
  }
  
  this.textWatch = new Watchable(text,DEFAULT_TEXT);
  this.urlWatch = new Watchable(url,DEFAULT_URL);
  this.xWatch = new Watchable(x,MIDDLE_COORDINATE);
  this.yWatch = new Watchable(y,MIDDLE_COORDINATE);
  this.viewWatch = new Watchable(view,null);
  this.clientWatch = new Watchable(client,null);
  this.subjectEdgeCountWatch = new Watchable(subjectEdgeCount,0);
  this.predicateEdgeCountWatch = new Watchable(predicateEdgeCount,0);
  this.objectEdgeCountWatch = new Watchable(objectEdgeCount,0);
  
  function updateNodeCounts ( oldView, newView ) {
    if ( oldView !== null ) {
      oldView.nodeCountWatch.decrement();
    }
    if ( newView !== null ) {
      newView.nodeCountWatch.increment();
    }
  }// end of function updateNodeCounts
  this.viewWatch.addChangeCallback(updateNodeCounts);
  
  this.cleanup = function () {
    // Deattach from the view.
    this.viewWatch.set(null);
  };// end of this.cleanup
  
}// end of function Node

function Edge ( id, subject, predicate, object, view ) {
  
  this.getId = function () {
    return id;
  };// end of this.getId
  
  this.subjectWatch = new Watchable(subject,null);
  this.predicateWatch = new Watchable(predicate,null);
  this.objectWatch = new Watchable(object,null);
  this.viewWatch = new Watchable(view,null);
  
  if ( subject !== null ) {
    subject.subjectEdgeCountWatch.increment();
  }
  
  function updateSubjectEdgeCounts ( oldNode, newNode ) {
    if ( oldNode !== null ) {
      oldNode.subjectEdgeCountWatch.decrement();
    }
    if ( newNode !== null ) {
      newNode.subjectEdgeCountWatch.increment();
    }
  }// end of function updateSubjectEdgeCounts
  this.subjectWatch.addChangeCallback(updateSubjectEdgeCounts);
  
  if ( predicate !== null ) {
    predicate.predicateEdgeCountWatch.increment();
  }
  
  function updatePredicateEdgeCounts ( oldNode, newNode ) {
    if ( oldNode !== null ) {
      oldNode.predicateEdgeCountWatch.decrement();
    }
    if ( newNode !== null ) {
      newNode.predicateEdgeCountWatch.increment();
    }
  }// end of function updatePredicateEdgeCounts
  this.predicateWatch.addChangeCallback(updatePredicateEdgeCounts);
  
  if ( object !== null ) {
    object.objectEdgeCountWatch.increment();
  }
  
  function updateObjectEdgeCounts ( oldNode, newNode ) {
    if ( oldNode !== null ) {
      oldNode.objectEdgeCountWatch.decrement();
    }
    if ( newNode !== null ) {
      newNode.objectEdgeCountWatch.increment();
    }
  }// end of function updateObjectEdgeCounts
  this.objectWatch.addChangeCallback(updateObjectEdgeCounts);
  
  if ( view !== null ) {
    view.edgeCountWatch.increment();
  }
  
  function updateEdgeCounts ( oldView, newView ) {
    if ( oldView !== null ) {
      oldView.edgeCountWatch.decrement();
    }
    if ( newView !== null ) {
      newView.edgeCountWatch.increment();
    }
  }// end of function updateEdgeCounts
  this.viewWatch.addChangeCallback(updateEdgeCounts);
  
  this.cleanup = function () {
    // Deattach from the view and from the nodes.
    this.subjectWatch.set(null);
    this.predicateWatch.set(null);
    this.objectWatch.set(null);
    this.viewWatch.set(null);
  };// end of this.cleanup
  
}// end of function Edge

// We use these objects to map a unique ID to each object.
// Messages from the server refer to objects by ID.
// get called with a null ID returns a null object.
// has called with a null ID returns false.
// Note that the objects being stored should have a cleanup method we can call before removal that takes in no arguments.
function WatchableMap () {
  
  var map = {};
  
  var createCallbacks = [];
  var deleteCallbacks = [];
  
  this.create = function ( id, item ) {
    map[id] = item;
    function callCallback ( callback ) {
      callback(id,item);
    }// end of function callCallback
    createCallbacks.forEach(callCallback);
  };// end of this.create
  
  this.delete = function ( id ) {
    var item = map[id];
    delete map[id];
    // We call the callbacks after deleting the item from the map so that they can see what keys remain.
    // We call them before calling cleanup so that the object is still in tact when they get it.
    function callCallback ( callback ) {
      callback(id,item);
    }// end of function callCallback
    deleteCallbacks.forEach(callCallback);
    item.cleanup();
  };// end of this.delete
  
  this.get = function ( id ) {
    if ( id === null ) {
      return null;
    }
    return map[id];
  };// end of this.get
  
  this.has = function ( id ) {
    if ( id === null ) {
      return false;
    }
    return map.hasOwnProperty(id);
  };// end of this.has
  
  // Note that we call callback with both ID and item as arguments.
  this.forEach = function ( callback ) {
    var item = null;
    for ( var id in map ) {
      callback(id, map[id]);
    }// end of for through IDs
  };// end of this.forEach
  
  this.addCreateCallback = function ( callback ) {
    createCallbacks.push(callback);
  };// end of this.addCreateCallback
  
  this.removeCreateCallback = function ( callback ) {
    createCallbacks.splice( createCallbacks.indexOf(callback), 1 );
  };// end of this.removeCreateCallback
  
  this.addDeleteCallback = function ( callback ) {
    deleteCallbacks.push(callback);
  };// end of this.addDeleteCallback
  
  this.removeDeleteCallback = function ( callback ) {
    deleteCallbacks.splice( deleteCallbacks.indexOf(callback), 1 );
  };// end of this.removeDeleteCallback
  
}// end of function WatchableMap

var viewMap = new WatchableMap();
var clientMap = new WatchableMap();
var nodeMap = new WatchableMap();
var edgeMap = new WatchableMap();

var myClientId = null;// the client ID of the local user

var setMyClientIdCallbacks = [];

function setMyClientId ( newId ) {
  var oldId = myClientId;
  myClientId = newId;
  function callCallback( callback ) {
    callback(oldId,newId);
  }// end of function callCallback
  setMyClientIdCallbacks.forEach(callCallback);
}// end of function setMyClientId

// action validator layer:

// These functions validate an action against certain rules regarding changes to the model:
// 1. To be created, something must not already exist.
//    Specifically, an object of the same type with the same ID must not exist.
// 2. To be updated, something must already exist.
//    Specifically, an object of the correct type with the correct ID must exist.
// 3. To be deleted, something must already exist.
// 4. To be referenced by another object, something must already exist.
// 5. A client can attach to a node if and only if no other client is attached.
// 6. A view must have no attached clients, no nodes, and no edges when we delete it.
// 7. A node must not participate in any edges before when we delete it.
// 8. Client user name must be a string.
// 9. Node text must be a string.
// 9b. Node URL must be a string.
// 10. Node x and y coordinates must be numbers such that 0 <= x < 1 and 0 <= y < 1.
// 11. A client does not need to be attached to a specific view.
// 12. A node does need to belong to a specific view.
// 13. An edge does need to belong to a specific view.
// 14. The creator of a node becomes its current client.
// 15. At other times, a node does not need to have a client.
// 16. A node can only be updated by its current client.
// 17. A node can only be deleted by its current client.
// 18. The creator of an edge must be the current client of all three component nodes.
// 19. Only the current client of all three component nodes of an edge can delete it.
// 20. A node is associated with a single view from its creation until its deletion.
// 21. An edge is associated with a single view from its creation until its deletion.
// 22. An edge is associated with a single triple of nodes from its creation until its deletion.
// 23. All three nodes in an edge must belong to the same view.
// 24. The edge itself belongs to the same view as its component nodes.
// 25. Since clients may disconnect unexpectedly, we cannot impose any conditions on deletion of a client.
// 26. No IDs can be the empty string.

// Each checking function expects a ValidationResult object, on which it performs the following actions:
// 1. Set VALID to false if the statement is false.
// 2. Add the reason why the statement is false to REASONS.

function ValidationResult () {
  
  var invalid = false;// boolean indicating whether any assertions checked so far are true
  var reasons = [];// array of strings giving telling which assertions were true
  
  this.invalidateIfTrue = function ( value, reason ) {
    if ( value ) {
      invalid = true;
      reasons.push(reason);
    }
  };// end of this.requireTrue
  
  this.getInvalid = function () {
    return invalid;
  };// end of this.getInvalid
  
  this.getReasons = function () {
    return reasons;
  };// end of this.getReasons
  
}// end of function ValidationResult

// The following are helper functions perform checks that occur repeatedly in the functions that map to actions.

// empty string test:
function idIsNotEmptyString ( id, result ) {
  result.invalidateIfTrue( id === "", "ID is the empty string." );
  return result;
}// end of function idIsNotEmptyString

// existence tests:

function thingExists (typeName, idMap, id, result) {
  result.invalidateIfTrue( !idMap.has(id), typeName+" with ID "+id+" does not exist." );
  return result;
}// end of function thingExists

function thingDoesNotExist (typeName, idMap, id, result) {
  result.invalidateIfTrue( idMap.has(id), typeName+" with ID "+id+" already exists." );
  return result;
}// end of function thingDoesNotExist

function viewExists ( id, result ) {
  thingExists("View", viewMap, id, result);
  return result;
}// end of function viewExists

function viewDoesNotExist ( id, result ) {
  thingDoesNotExist("View", viewMap, id, result);
  return result;
}// end of function viewDoesNotExist

function clientExists ( id, result ) {
  thingExists("Client", clientMap, id, result);
  return result;
}// end of function clientExists

function clientDoesNotExist ( id, result ) {
  thingDoesNotExist("Client", clientMap, id, result);
  return result;
}// end of function clientDoesNotExist

function nodeExists ( id, result ) {
  thingExists("Node", nodeMap, id, result);
  return result;
}// end of function nodeExists

function nodeDoesNotExist ( id, result ) {
  thingDoesNotExist("Node", nodeMap, id, result);
  return result;
}// end of function nodeDoesNotExist

function edgeExists ( id, result ) {
  thingExists("Edge", edgeMap, id, result);
  return result;
}// end of function edgeExists

function edgeDoesNotExist ( id, result ) {
  thingDoesNotExist("Edge", edgeMap, id, result);
  return result;
}// end of function edgeDoesNotExist

// Note that the helper function tests below do not call the existence tests,
// because we assume that a caller interested in validating existence can call them before calling these.
// We do, however, check for existence in case the caller does pass in a bad ID.

// Freedom for deletion tests:

function viewIsEmpty ( id, result ) {
  if ( !viewMap.has(id) ) {
    console.log("Warning: viewIsEmpty received unrecognized view ID "+id+".");
    return result;
  }
  var view = viewMap.get(id);
  var clientCount = view.clientCountWatch.get();
  var nodeCount = view.nodeCountWatch.get();
  var edgeCount = view.edgeCountWatch.get();
  var invalid = (clientCount > 0)||(nodeCount > 0)||(edgeCount > 0);
  var reason = "View "+id+" is non-empty, has "+clientCount+" clients editing, "+nodeCount+" nodes, "+edgeCount+" edges.";
  result.invalidateIfTrue( invalid, reason );
  return result;
}// end of function viewIsEmpty

function nodeDoesNotParticipateInEdges ( id, result ) {
  if ( !nodeMap.has(id) ) {
    console.log("Warning: nodeDoesNotParticipateInEdges received unrecognized node ID "+id+".");
    return result;
  }
  var node = nodeMap.get(id);
  subjectCount = node.subjectEdgeCountWatch.get();
  predicateCount = node.predicateEdgeCountWatch.get();
  objectCount = node.objectEdgeCountWatch.get();
  result.invalidateIfTrue(subjectCount > 0, "Node "+id+" is the subject in "+subjectCount+" edges.");
  result.invalidateIfTrue(predicateCount > 0, "Node "+id+" is the predicate in "+predicateCount+" edges.");
  result.invalidateIfTrue(objectCount > 0, "Node "+id+" is the object in "+objectCount+" edges.");
  return result;
}// end of nodeDoesNotParticipateInEdges

// Node-client attachment test:

function nodeHasThisClient ( nodeId, clientId, result ) {
  if ( !nodeMap.has(nodeId) ) {
    console.log("Warning: nodeHasThisClient received unrecognized node ID "+nodeId+".");
    return result;
  }
  if ( !clientMap.has(clientId) ) {
    console.log("Warning: nodeHasThisClient received unrecognized client ID "+clientId+".");
    return result;
  }
  var nodeClient = nodeMap.get(nodeId).clientWatch.get();
  var client = clientMap.get(clientId);
  result.invalidateIfTrue( client !== nodeClient, "Node "+nodeId+" does not have client "+clientId+"." );
  return result;
}// end of function nodeHasThisClient

function nodeHasThisClientOrNone ( nodeId, clientId, result ) {
  if ( !nodeMap.has(nodeId) ) {
    console.log("Warning: nodeHasThisClientOrNone received unrecognized node ID "+nodeId+".");
    return result;
  }
  if ( !clientMap.has(clientId) ) {
    console.log("Warning: nodeHasThisClientOrNone received unrecognized client ID "+clientId+".");
    return result;
  }
  var nodeClient = nodeMap.get(nodeId).clientWatch.get();
  var client = clientMap.get(clientId);
  result.invalidateIfTrue( (nodeClient !== client) && (nodeClient !== null), "Node "+nodeId+" has a client other than "+clientId+"." );
  return result;
}// end of function nodeHasThisClientOrNone

function edgeNodesHaveThisClient ( edgeId, clientId, result ) {
  if ( !edgeMap.has(edgeId) ) {
    console.log("Warning: edgeNodesHaveThisClient received unrecognized edge ID "+edgeId+".");
    return result;
  }
  if ( !clientMap.has(clientId) ) {
    console.log("Warning: edgeNodesHaveThisClient received unrecognized client ID "+clientId+".");
    return result;
  }
  var edge = edgeMap.get(edgeId);
  var client = clientMap.get(clientId);
  // For a valid edge, the component nodes should never be null.
  var subjectClient = edge.subjectWatch.get().clientWatch.get();
  var predicateClient = edge.predicateWatch.get().clientWatch.get();
  var objectClient = edge.objectWatch.get().clientWatch.get();
  result.invalidateIfTrue( client !== subjectClient, "The subject node of edge "+edgeId+" does not have client "+clientId+"." );
  result.invalidateIfTrue( client !== predicateClient, "The predicate node of edge "+edgeId+" does not have client "+clientId+"." );
  result.invalidateIfTrue( client !== objectClient, "The object node of edge "+edgeId+" does not have client "+clientId+"." );
  return result;
}// end of function edgeNodesHaveThisClient

// node-view attachment test:

function nodesHaveSameView ( nodeIds, result ) {
  var firstView = null;
  if ( nodeIds.length === 0 ) {
    return;// There is nothing to do.
  }
  var firstId = nodeIds[0];
  if ( !nodeMap.has(firstId) ) {
    console.log("Warning: nodesHaveSameView received unrecognized node ID "+firstId+".");
    return;
  }
  var firstView = nodeMap.get(firstId).viewWatch.get();
  function nodeHasSameViewAsFirst (nodeId) {
    if ( !nodeMap.has(nodeId) ) {
      console.log("Warning: nodesHaveSameView received unrecognized node ID "+nodeId+".");
      return;
    }
    result.invalidateIfTrue( nodeMap.get(nodeId).viewWatch.get() !== firstView, "Node "+nodeId+" is not in the same view as "+firstId+"." );
  }// end of function nodeHasSameViewAsFirst
  nodeIds.forEach(nodeHasSameViewAsFirst);
  return result;
}// end of function nodesHaveSameView

// valid value tests:

function argIsString ( argName, arg, result ) {
  // If we wanted to test for String objects, we could use userName instanceof String,
  // but we should not be using those.
  result.invalidateIfTrue( typeof arg !== 'string', argName+" has type "+(typeof arg)+", not string." );
  return result;
}// end of function argIsString

function argIsCoordinateInRange ( argName, arg, result ) {
  var argNum = Number(arg);
  var argNaN = isNaN(argNum);
  result.invalidateIfTrue(argNaN, argName+" value "+arg+" is not a valid number.");
  if ( !argNaN ) {
    result.invalidateIfTrue(argNum < MIN_COORDINATE, argName+" value "+argNum+" is less than "+MIN_COORDINATE+".");
    result.invalidateIfTrue(MAX_COORDINATE < argNum, argName+" value "+argNum+" is greater than "+MAX_COORDINATE+".");
  }
  return result;
}// end of function argIsCoordinateInRange

// The following functions map directly to actions.

// view:

function canCreateView ( viewId, result ) {
  // 26. No IDs can be the empty string.
  idIsNotEmptyString (viewId, result);
  if ( result.getInvalid() ) {
    return result;
  }
  // 1. To be created, something must not already exist.
  viewDoesNotExist(viewId, result);
  return result;
}// end of function canCreateView

function canDeleteView ( viewId, result ) {
  // 3. To be deleted, something must already exist.
  viewExists(viewId, result);
  if ( result.getInvalid() ) {
    return result;
  }
  // 6. A view must have no attached clients, no nodes, and no edges when we delete it.
  viewIsEmpty(viewId, result);
  return result;
}// end of function canDeleteView

// client:

function canCreateClient ( clientId, userName, viewId, result ) {
  // 26. No IDs can be the empty string.
  idIsNotEmptyString (clientId, result);
  if ( result.getInvalid() ) {
    return result;
  }
  // 1. To be created, something must not already exist.
  clientDoesNotExist(clientId, result);
  if ( result.getInvalid() ) {
    return result;
  }
  // 8. Client user name must be a string.
  argIsString("userName", userName, result);
  if ( result.getInvalid() ) {
    return result;
  }
  // 11. A client does not need to be attached to a specific view.
  if ( viewId !== null ) {
    // 4. To be referenced by another object, something must already exist.
    viewExists(viewId, result);
  }
  return result;
}// end of function canCreateClient

function canUpdateClientUserName ( clientId, userName, result ) {
  // 2. To be updated, something must already exist.
  clientExists(clientId, result);
  if ( result.getInvalid() ) {
    return result;
  }
  // 8. Client user name must be a string.
  argIsString("userName", userName, result);
  return result;
}// end of function canUpdateClientUserName

function canUpdateClientView ( clientId, viewId, result ) {
  // 2. To be updated, something must already exist.
  clientExists(clientId, result);
  if ( result.getInvalid() ) {
    return result;
  }
  // 11. A client does not need to be attached to a specific view.
  if ( viewId !== null ) {
    // 4. To be referenced by another object, something must already exist.
    viewExists(viewId, result);
  }
  return result;
}// end of function canUpdateClientView

function canDeleteClient ( clientId, result ) {
  // 3. To be deleted, something must already exist.
  clientExists(clientId, result);
  return result;
}// end of function canDeleteClient

// node:

function canCreateNode ( nodeId, text, url, x, y, viewId, clientId, result ) {
  // 26. No IDs can be the empty string.
  idIsNotEmptyString (nodeId, result);
  if ( result.getInvalid() ) {
    return result;
  }
  // 1. To be created, something must not already exist.
  nodeDoesNotExist(nodeId, result);
  if ( result.getInvalid() ) {
    return result;
  }
  // 9. Node text must be a string.
  argIsString("node text", text, result);
  if ( result.getInvalid() ) {
    return result;
  }
  // 9b. Node URL must be a string.
  argIsString("node URL", url, result);
  if ( result.getInvalid() ) {
    return result;
  }
  // 10. Node x and y coordinates must be numbers such that 0 <= x < 1 and 0 <= y < 1.
  argIsCoordinateInRange("node x", x, result);
  if ( result.getInvalid() ) {
    return result;
  }
  argIsCoordinateInRange("node y", y, result);
  if ( result.getInvalid() ) {
    return result;
  }
  // 12. A node does need to belong to a specific view.
  // 4. To be referenced by another object, something must already exist.
  viewExists(viewId, result);
  if ( result.getInvalid() ) {
    return result;
  }
  // 14. The creator of a node becomes its current client.
  // 4. To be referenced by another object, something must already exist.
  clientExists(clientId, result);
  return result;
}// end of function canCreateNode

function canUpdateNodeText ( nodeId, text, clientId, result ) {
  // 2. To be updated, something must already exist.
  nodeExists(nodeId, result);
  if ( result.getInvalid() ) {
    return result;
  }
  // 9. Node text must be a string.
  argIsString("node text", text, result);
  if ( result.getInvalid() ) {
    return result;
  }
  // 4. To be referenced by another object, something must already exist.
  clientExists(clientId, result);
  if ( result.getInvalid() ) {
    return result;
  }
  // 16. A node can only be updated by its current client.
  nodeHasThisClient(nodeId, clientId, result);
  return result;
}// end of function canUpdateNodeText

function canUpdateNodeUrl ( nodeId, url, clientId, result ) {
  // 2. To be updated, something must already exist.
  nodeExists(nodeId, result);
  if ( result.getInvalid() ) {
    return result;
  }
  // 9b. Node URL must be a string.
  argIsString("node URL", url, result);
  if ( result.getInvalid() ) {
    return result;
  }
  // 4. To be referenced by another object, something must already exist.
  clientExists(clientId, result);
  if ( result.getInvalid() ) {
    return result;
  }
  // 16. A node can only be updated by its current client.
  nodeHasThisClient(nodeId, clientId, result);
  return result;
}// end of function canUpdateNodeUrl

function canUpdateNodeX ( nodeId, x, clientId, result ) {
  // 2. To be updated, something must already exist.
  nodeExists(nodeId, result);
  if ( result.getInvalid() ) {
    return result;
  }
  // 10. Node x and y coordinates must be numbers such that 0 <= x < 1 and 0 <= y < 1.
  argIsCoordinateInRange("node x", x, result);
  if ( result.getInvalid() ) {
    return result;
  }
  // 4. To be referenced by another object, something must already exist.
  clientExists(clientId, result);
  if ( result.getInvalid() ) {
    return result;
  }
  // 16. A node can only be updated by its current client.
  nodeHasThisClient(nodeId, clientId, result);
  return result;
}// end of function canUpdateNodeX

function canUpdateNodeY ( nodeId, y, clientId, result ) {
  // 2. To be updated, something must already exist.
  nodeExists(nodeId, result);
  if ( result.getInvalid() ) {
    return result;
  }
  // 10. Node x and y coordinates must be numbers such that 0 <= x < 1 and 0 <= y < 1.
  argIsCoordinateInRange("node y", y, result);
  if ( result.getInvalid() ) {
    return result;
  }
  // 4. To be referenced by another object, something must already exist.
  clientExists(clientId, result);
  if ( result.getInvalid() ) {
    return result;
  }
  // 16. A node can only be updated by its current client.
  nodeHasThisClient(nodeId, clientId, result);
  return result;
}// end of function canUpdateNodeText

// 20. A node is associated with a single view from its creation until its deletion.

function canUpdateNodeClient ( nodeId, clientId, result ) {
  // 2. To be updated, something must already exist.
  nodeExists(nodeId, result);
  if ( result.getInvalid() ) {
    return result;
  }
  // 4. To be referenced by another object, something must already exist.
  clientExists(clientId, result);
  if ( result.getInvalid() ) {
    return result;
  }
  // 5. A client can attach to a node if and only if no other client is attached.
  nodeHasThisClientOrNone(nodeId, clientId, result);
  return result;
}// end of function canUpdateNodeText

function canDeleteNode ( nodeId, clientId, result ) {
  // 3. To be deleted, something must already exist.
  nodeExists(nodeId, result);
  if ( result.getInvalid() ) {
    return result;
  }
  // 4. To be referenced by another object, something must already exist.
  clientExists(clientId, result);
  if ( result.getInvalid() ) {
    return result;
  }
  // 17. A node can only be deleted by its current client.
  nodeHasThisClient(nodeId, clientId, result);
  if ( result.getInvalid() ) {
    return result;
  }
  // 7. A node must not participate in any edges before when we delete it.
  nodeDoesNotParticipateInEdges(nodeId, result);
  return result;
}// end of function canDeleteNode

// edge:

function canCreateEdge ( edgeId, subjectId, predicateId, objectId, clientId, result ) {
  // 26. No IDs can be the empty string.
  idIsNotEmptyString (edgeId, result);
  if ( result.getInvalid() ) {
    return result;
  }
  // 1. To be created, something must not already exist.
  edgeDoesNotExist(edgeId, result);
  if ( result.getInvalid() ) {
    return result;
  }
  // 4. To be referenced by another object, something must already exist.
  nodeExists(subjectId, result);
  if ( result.getInvalid() ) {
    return result;
  }
  nodeExists(predicateId, result);
  if ( result.getInvalid() ) {
    return result;
  }
  nodeExists(objectId, result);
  if ( result.getInvalid() ) {
    return result;
  }
  clientExists(clientId, result);
  if ( result.getInvalid() ) {
    return result;
  }
  // 18. The creator of an edge must be the current client of all three component nodes.
  nodeHasThisClient(subjectId, clientId, result);
  if ( result.getInvalid() ) {
    return result;
  }
  nodeHasThisClient(predicateId, clientId, result);
  if ( result.getInvalid() ) {
    return result;
  }
  nodeHasThisClient(objectId, clientId, result);
  if ( result.getInvalid() ) {
    return result;
  }
  // 23. All three nodes in an edge must belong to the same view.
  nodesHaveSameView([subjectId,predicateId,objectId], result);
  return result;
}// end of function canCreateEdge

// 22. An edge is associated with a single triple of nodes from its creation until its deletion.

// 21. An edge is associated with a single view from its creation until its deletion.

function canDeleteEdge ( edgeId, clientId, result ) {
  // 3. To be deleted, something must already exist.
  edgeExists(edgeId, result);
  if ( result.getInvalid() ) {
    return result;
  }
  // 4. To be referenced by another object, something must already exist.
  clientExists(clientId, result);
  if ( result.getInvalid() ) {
    return result;
  }
  // 19. Only the current client of all three component nodes of an edge can delete it.
  edgeNodesHaveThisClient(edgeId, clientId, result);
  return result;
}// end of function canCreateEdge

// Action received callbacks:

// Call the callbacks from each array when the corresponding action has been received.
// The callback should take the same arguments as the receive... function.

// view:
var callbacksCreateView = [];
var callbacksDeleteView = [];
// client:
var callbacksCreateClient = [];
var callbacksUpdateClientUserName = [];
var callbacksUpdateClientView = [];
var callbacksDeleteClient = [];
// node:
var callbacksCreateNode = [];
var callbacksUpdateNodeText = [];
var callbacksUpdateNodeUrl = [];
var callbacksUpdateNodeX = [];
var callbacksUpdateNodeY = [];
var callbacksUpdateNodeClient = [];
var callbacksDeleteNode = [];
// edge:
var callbacksCreateEdge = [];
var callbacksDeleteEdge = [];

// model mutator layer:

// We update the model when we receive a message from the server,
// not when the local client sends the message to the server.
// These functions retrieve objects based on the IDs provided in messages.
// A null value for the ID is translated to a null value for the object. 
// If the change is not valid, return without making any changes to the model.

// view:

function receiveCreateView ( viewId ) {  
  var result = canCreateView( viewId, new ValidationResult() );
  if ( result.getInvalid() ) {
    console.log( "invalid call to receiveCreateView: ", result.getReasons() );
    return;
  }
  viewMap.create( viewId, new View(viewId,0,0,0 ) );
  function callCallback ( callback ) {
    callback( viewId );
  }// end of function callCallback
  callbacksCreateView.forEach( callCallback );
}// end of function receiveCreateView

function receiveDeleteView ( viewId ) {
  var result = canDeleteView( viewId, new ValidationResult() );
  if ( result.getInvalid() ) {
    console.log( "invalid call to receiveDeleteView: ", result.getReasons() );
    return;
  }
  viewMap.delete(viewId);
  function callCallback ( callback ) {
    callback( viewId );
  }// end of function callCallback
  callbacksDeleteView.forEach( callCallback );
}// end of function receiveDeleteView

// client:

function receiveCreateClient ( clientId, userName, viewId ) {
  var result = canCreateClient( clientId, userName, viewId, new ValidationResult() );
  if ( result.getInvalid() ) {
    console.log( "invalid call to receiveCreateClient: ", result.getReasons() );
    return;
  }
  clientMap.create(  clientId,  new Client( clientId, userName, viewMap.get(viewId) )  );
  function callCallback ( callback ) {
    callback( clientId, userName, viewId );
  }// end of function callCallback
  callbacksCreateClient.forEach( callCallback );
}// end of function receiveCreateClient

function receiveUpdateClientUserName ( clientId, userName ) {
  var result = canUpdateClientUserName( clientId, userName, new ValidationResult() );
  if ( result.getInvalid() ) {
    console.log( "invalid call to receiveUpdateClientUserName: ", result.getReasons() );
    return;
  }
  clientMap.get(clientId).userNameWatch.set(userName);
  function callCallback ( callback ) {
    callback( clientId, userName );
  }// end of function callCallback
  callbacksUpdateClientUserName.forEach( callCallback );
}// end of function receiveUpdateClientUserName

function receiveUpdateClientView ( clientId, viewId ) {
  var result = canUpdateClientView( clientId, viewId, new ValidationResult() );
  if ( result.getInvalid() ) {
    console.log( "invalid call to receiveUpdateClientView: ", result.getReasons() );
    return;
  }
  clientMap.get(clientId).viewWatch.set( viewMap.get(viewId) );
  function callCallback ( callback ) {
    callback( clientId, viewId );
  }// end of function callCallback
  callbacksUpdateClientView.forEach( callCallback );
}// end of function receiveUpdateClientView

function receiveDeleteClient ( clientId ) {
  var result = canDeleteClient( clientId, new ValidationResult() );
  if ( result.getInvalid() ) {
    console.log( "invalid call to receiveDeleteClient: ", result.getReasons() );
    return;
  }
  // Since the client may exit without deattaching from nodes, we need to perform this cleanup.
  // Since client does not contain references to the nodes to which it is attached,
  // we need to search through all the nodes for them.
  // Since this involves lookup in the ID->Object map, we do it here rather than in Client's own cleanup method.
  var client = clientMap.get(clientId);
  function detachDeletedClient ( nodeId, node ) {
    if ( node.clientWatch.get() === client ) {
      node.clientWatch.set(null);
    }
  }// end of function detachDeletedClient
  nodeMap.forEach(detachDeletedClient);
  // The regular cleanup method called in clientMap.delete takes care of views.
  clientMap.delete(clientId);
  function callCallback ( callback ) {
    callback( clientId );
  }// end of function callCallback
  callbacksDeleteClient.forEach( callCallback );
}// end of function receiveDeleteClient

// node:

function receiveCreateNode ( nodeId, text, url, x, y, viewId, clientId ) {
  var result = canCreateNode( nodeId, text, url, x, y, viewId, clientId, new ValidationResult() );
  if ( result.getInvalid() ) {
    console.log( "invalid call to receiveCreateNode: ", result.getReasons() );
    return;
  }
  nodeMap.create(  nodeId,  new Node( nodeId, text, url, Number(x), Number(y), viewMap.get(viewId), clientMap.get(clientId), 0, 0, 0 )  );
  function callCallback ( callback ) {
    callback( nodeId, text, x, y, viewId, clientId );
  }// end of function callCallback
  callbacksCreateNode.forEach( callCallback );
}// end of function receiveCreateNode

function receiveUpdateNodeText ( nodeId, text, clientId ) {
  var result = canUpdateNodeText( nodeId, text, clientId, new ValidationResult() );
  if ( result.getInvalid() ) {
    console.log( "invalid call to receiveUpdateNodeText: ", result.getReasons() );
    return;
  }
  nodeMap.get(nodeId).textWatch.set(text);
  function callCallback ( callback ) {
    callback( nodeId, text, clientId );
  }// end of function callCallback
  callbacksUpdateNodeText.forEach( callCallback );
}// end of function receiveUpdateNodeText

function receiveUpdateNodeUrl ( nodeId, url, clientId ) {
  var result = canUpdateNodeUrl( nodeId, url, clientId, new ValidationResult() );
  if ( result.getInvalid() ) {
    console.log( "invalid call to receiveUpdateNodeUrl: ", result.getReasons() );
    return;
  }
  nodeMap.get(nodeId).urlWatch.set(url);
  function callCallback ( callback ) {
    callback( nodeId, url, clientId );
  }// end of function callCallback
  callbacksUpdateNodeUrl.forEach( callCallback );
}// end of function receiveUpdateNodeUrl

function receiveUpdateNodeX ( nodeId, x, clientId ) {
  var result = canUpdateNodeX( nodeId, x, clientId, new ValidationResult() );
  if ( result.getInvalid() ) {
    console.log( "invalid call to receiveUpdateNodeX: ", result.getReasons() );
    return;
  }
  nodeMap.get(nodeId).xWatch.set( Number(x) );
  function callCallback ( callback ) {
    callback( nodeId, x, clientId );
  }// end of function callCallback
  callbacksUpdateNodeX.forEach( callCallback );
}// end of function receiveUpdateNodeX

function receiveUpdateNodeY ( nodeId, y, clientId ) {
  var result = canUpdateNodeY( nodeId, y, clientId, new ValidationResult() );
  if ( result.getInvalid() ) {
    console.log( "invalid call to receiveUpdateNodeY: ", result.getReasons() );
    return;
  }
  nodeMap.get(nodeId).yWatch.set( Number(y) );
  function callCallback ( callback ) {
    callback( nodeId, y, clientId );
  }// end of function callCallback
  callbacksUpdateNodeY.forEach( callCallback );
}// end of function receiveUpdateNodeY

function receiveUpdateNodeClient ( nodeId, clientId ) {
  var result = canUpdateNodeClient( nodeId, clientId, new ValidationResult() );
  if ( result.getInvalid() ) {
    console.log( "invalid call to receiveUpdateNodeClient: ", result.getReasons() );
    return;
  }
  // For this call to be valid, one of two cases must be true:
  var node = nodeMap.get(nodeId);
  var client = clientMap.get(clientId);
  if ( node.clientWatch.get() === client ) {
    // 1. This client is already the current client, so we deattach it.
    node.clientWatch.set(null);
  }
  else {
    // 2. There is no current client, so we attach this client.
    node.clientWatch.set(client);
  }
  function callCallback ( callback ) {
    callback( nodeId, clientId );
  }// end of function callCallback
  callbacksUpdateNodeClient.forEach( callCallback );
}// end of function receiveUpdateNodeClient

function receiveDeleteNode ( nodeId, clientId ) {
  var result = canDeleteNode( nodeId, clientId, new ValidationResult() );
  if ( result.getInvalid() ) {
    console.log( "invalid call to receiveDeleteNode: ", result.getReasons() );
    return;
  }
  nodeMap.delete(nodeId);
  function callCallback ( callback ) {
    callback( nodeId, clientId );
  }// end of function callCallback
  callbacksDeleteNode.forEach( callCallback );
}// end of function receiveDeleteNode

// edge:

function receiveCreateEdge ( edgeId, subjectId, predicateId, objectId, clientId ) {
  var result = canCreateEdge( edgeId, subjectId, predicateId, objectId, clientId, new ValidationResult() );
  if ( result.getInvalid() ) {
    console.log( "invalid call to receiveCreateEdge: ", result.getReasons() );
    return;
  }
  var subject = nodeMap.get(subjectId);
  var predicate = nodeMap.get(predicateId);
  var object = nodeMap.get(objectId);
  var view = null;
  if ( subject !== null ) {
    view = subject.viewWatch.get();
  }
  edgeMap.create(  edgeId,  new Edge( edgeId, subject, predicate, object, view )  );
  function callCallback ( callback ) {
    callback( edgeId, subjectId, predicateId, objectId, clientId );
  }// end of function callCallback
  callbacksCreateEdge.forEach( callCallback );
}// end of function receiveCreateEdge

function receiveDeleteEdge ( edgeId, clientId ) {
  var result = canDeleteEdge( edgeId, clientId, new ValidationResult() );
  if ( result.getInvalid() ) {
    console.log( "invalid call to receiveDeleteEdge: ", result.getReasons() );
    return;
  }
  edgeMap.delete(edgeId);
  function callCallback ( callback ) {
    callback( edgeId, clientId );
  }// end of function callCallback
  callbacksDeleteEdge.forEach( callCallback );
}// end of function receiveCreateEdge

// message format: object with string keys defined below

// Every message must have an action.

// filterFor is a special action sent to the server, never received.
// It tells the server to only send this client messages that contain the specified key-value pairs.
// Besides action, the message object must have exactly one other key-value pair, filter: {.}, a filter object.
// A filter object has exactly one key-value pair:
// and: [...], or: [...], not: {.}, or any one of the key-value pairs a message can have.
// {.} must itself be a valid filter object.
// [...] must be an array of valid filter objects.
// The server tests recursively the different requirements and only sends to the client messages that meet them.

// getBacklog is a special action that the client sends to the server
// and receives back with an array of all messages including and after the start index.

// getMyClientId is a special action that the client sends to the server
// and receives back with the ID the server assigned to the local client.

// Each other value of action maps to one or more of the receive... functions above.
// The server relays sent messages with these actions to all clients that have not set filters to block them.
// key:
var ACTION = 'action';
// values:
var SET_FILTER = 'setFilter';// must include a single other key-value pair
var GET_BACKLOG = 'getBacklog';// can include startIndex when sent, must include backlog when received
var GET_MY_CLIENT_ID = 'getMyClientId';// should not include any other key-value pairs
var CREATE_VIEW = 'createView';// must include viewId
var DELETE_VIEW = 'deleteView';// must include viewId
var CREATE_CLIENT = 'createClient';// must include clientId, userName, viewId
var UPDATE_CLIENT = 'updateClient';// must include clientId and userName or viewId
var DELETE_CLIENT = 'deleteClient';// must include clientId
var CREATE_NODE = 'createNode';// must include nodeId, text, url, x, y, viewId, and clientId
var UPDATE_NODE = 'updateNode';// must include nodeId, clientId, and text, url, x, or y
var DELETE_NODE = 'deleteNode';// must include nodeId and clientId
var CREATE_EDGE = 'createEdge';// must include edgeId, subjectId, predicateId, objectId, viewId, and clientId
var DELETE_EDGE = 'deleteEdge';// must include edgeId and clientId

// other keys:
var FILTER = 'filter';// value type Object
var START_INDEX = 'startIndex';// value type number (integer)
var BACKLOG = 'backlog';// value type array of message objects
var MY_CLIENT_ID = 'myClientId';// value type number (integer)
var VIEW_ID = 'viewId';// value type string
var CLIENT_ID = 'clientId';// value type number
var NODE_ID = 'nodeId';// value type string
var EDGE_ID = 'edgeId';// value type string
var USER_NAME = 'userName';// value type string
var TEXT = 'text';// value type string
var URL = 'url';// value type string
var X = 'x';// value type number
var Y = 'y';// value type number
var SUBJECT_ID = 'subjectId';// value type string
var PREDICATE_ID = 'predicateId';// value type string
var OBJECT_ID = 'objectId';// value type string
var SERVER_TIMESTAMP = 'serverTimestamp';// value type number (integer) -The server adds this when it receives the message.

// message interpreting layer:

// Log messages to this array.
// We want this array to resemble the one the server is keeping,
// so store everything except the special messages getBacklog and getMyClientId.
var loggedMessages = [];

// Call callbacks from this array when a logable message arrives.
var logMessageCallbacks = [];

// logable messages: Log all messages from the server except getBacklog messages.
// The individual messages in the backlog get logged separately instead.
function logMessage(message) {
  if ( message.hasOwnProperty(ACTION) ) {
    if ( message[ACTION] === GET_BACKLOG ) {
      return;
    }
    if ( message[ACTION] === GET_MY_CLIENT_ID ) {
      return;
    }
  }
  loggedMessages.push(message);
  function callCallback (callback) {
    callback(message);
  }// end of function callCallback
  logMessageCallbacks.forEach(callCallback);
}// end of function logMessage

function missingRequiredProperties(message, requiredProperties) {
  function isNotInMessage(property) {
    return !message.hasOwnProperty(property);
  }// end of function isInMessage
  var missingProperties = requiredProperties.filter(isNotInMessage);
  function outputMissingMessage(property) {
    console.log("onMessage called with message with "+ACTION+" "+message[ACTION]+" missing property "+property);
  }// end of function outputMissingMessage
  missingProperties.forEach(outputMissingMessage);
  if (missingProperties.length > 0) {
    console.log("message: ", message);
    return true;
  }
  else {
    return false;
  }
}// end of function hasRequiredProperties

function onMessageFromServer (message) {
  logMessage(message);
  // Use the action attribute to map the message to the appropriate change to the model.
  if ( !message.hasOwnProperty(ACTION) ) {
    console.log("onMessage called with message missing "+ACTION+" property: ", message);
    return;
  }
  switch ( message[ACTION] ) {
    case CREATE_VIEW:
      if ( missingRequiredProperties(message,[VIEW_ID]) ) {
        return;
      }
      receiveCreateView(message[VIEW_ID]);
    break;
    case DELETE_VIEW:
      if ( missingRequiredProperties(message,[VIEW_ID]) ) {
        return;
      }
      receiveDeleteView(message[VIEW_ID]);
    break;
    case CREATE_CLIENT:
      if ( missingRequiredProperties(message,[CLIENT_ID,USER_NAME]) ) {
        return;
      }
      // Since the client does not need to be attached to a view, allow omission of VIEW_ID.
      if ( !message.hasOwnProperty(VIEW_ID) ) {
        message[VIEW_ID] = null;
      }
      receiveCreateClient(message[CLIENT_ID],message[USER_NAME],message[VIEW_ID]);
    break;
    case UPDATE_CLIENT:
      if ( missingRequiredProperties(message,[CLIENT_ID]) ) {
        return;
      }
      if ( message.hasOwnProperty(USER_NAME) ) {
        receiveUpdateClientUserName(message[CLIENT_ID],message[USER_NAME]);
      }
      if ( message.hasOwnProperty(VIEW_ID) ) {
        receiveUpdateClientView(message[CLIENT_ID],message[VIEW_ID]);
      }
    break;
    case DELETE_CLIENT:
      if ( missingRequiredProperties(message,[CLIENT_ID]) ) {
        return;
      }
      receiveDeleteClient(message[CLIENT_ID]);
    break;
    case CREATE_NODE:
      if ( missingRequiredProperties(message,[NODE_ID,TEXT,URL,X,Y,VIEW_ID,CLIENT_ID]) ) {
        return;
      }
      receiveCreateNode(message[NODE_ID],message[TEXT],message[URL],message[X],message[Y],message[VIEW_ID],message[CLIENT_ID]);
    break;
    case UPDATE_NODE:
      if ( missingRequiredProperties(message,[NODE_ID,CLIENT_ID]) ) {
        return;
      }
      var hasText = message.hasOwnProperty(TEXT);
      var hasUrl = message.hasOwnProperty(URL);
      var hasX = message.hasOwnProperty(X);
      var hasY = message.hasOwnProperty(Y);
      if ( hasText ) {
        receiveUpdateNodeText(message[NODE_ID],message[TEXT],message[CLIENT_ID]);
      }
      if ( hasUrl ) {
        receiveUpdateNodeUrl(message[NODE_ID],message[URL],message[CLIENT_ID]);
      }
      if ( hasX ) {
        receiveUpdateNodeX(message[NODE_ID],message[X],message[CLIENT_ID]);
      }
      if ( hasY ) {
        receiveUpdateNodeY(message[NODE_ID],message[Y],message[CLIENT_ID]);
      }
      // If none of the optional properties are present, assume we are updating the attached client.
      if ( !(hasText||hasUrl||hasX||hasY) ) {
        receiveUpdateNodeClient(message[NODE_ID],message[CLIENT_ID]);
      }
    break;
    case DELETE_NODE:
      if ( missingRequiredProperties(message,[NODE_ID,CLIENT_ID]) ) {
        return;
      }
      receiveDeleteNode(message[NODE_ID],message[CLIENT_ID]);
    break;
    case CREATE_EDGE:
      if ( missingRequiredProperties(message,[EDGE_ID,SUBJECT_ID,PREDICATE_ID,OBJECT_ID,CLIENT_ID]) ) {
        return;
      }
      receiveCreateEdge(message[EDGE_ID],message[SUBJECT_ID],message[PREDICATE_ID],message[OBJECT_ID],message[CLIENT_ID]);
    break;
    case DELETE_EDGE:
      if ( missingRequiredProperties(message,[EDGE_ID,CLIENT_ID]) ) {
        return;
      }
      receiveDeleteEdge(message[EDGE_ID],message[CLIENT_ID]);
    break;
    case GET_BACKLOG:
      if ( missingRequiredProperties(message,[BACKLOG]) ) {
        return;
      }
      message[BACKLOG].forEach(onMessageFromServer);
    break;
    case GET_MY_CLIENT_ID:
      if ( missingRequiredProperties(message,[MY_CLIENT_ID]) ) {
        return;
      }
      setMyClientId( message[MY_CLIENT_ID] );
    break;
    default:
      console.log("onMessage called with unrecognized "+ACTION+" value "+message[ACTION]+": ", message);
      return;
  }// end of switch over action values
}// end of function onMessageFromServer

// message sending layer:

var primus = null;

function sendToServer(message) {
    //console.log("sending to server: ", message);
    primus.write(message);
};//end of function sendToServer

// message creation layer:

// Note that none of these include the client ID.
// The server adds the client ID when it receives the message.

// special:

function sendSetFilter ( filter ) {
  var message = {};
  message[ACTION] = SET_FILTER;
  message[FILTER] = filter;
  sendToServer(message);
}// end of function sendSetFilter

function sendGetBacklog ( startIndex ) {
  var message = {};
  message[ACTION] = GET_BACKLOG;
  message[START_INDEX] = startIndex;
  sendToServer(message);
}// end of function sendGetBacklog

function sendGetBacklogAfterLastReceived () {
  var message = {};
  message[ACTION] = GET_BACKLOG;
  message[START_INDEX] = loggedMessages.length;
  sendToServer(message);
}// end of function sendGetBacklog

function sendGetMyClientId () {
  var message = {};
  message[ACTION] = GET_MY_CLIENT_ID;
  sendToServer(message);
}// end of function sendGetBacklog

function sendSetupRequests () {
  sendGetBacklogAfterLastReceived();
  sendGetMyClientId();
}// end of function sendSetupRequests

// view:

function sendCreateView ( viewId ) {
  var message = {};
  message[ACTION] = CREATE_VIEW;
  message[VIEW_ID] = viewId;
  sendToServer(message);
}// end of function sendCreateView

function sendDeleteView ( viewId ) {
  var message = {};
  message[ACTION] = DELETE_VIEW;
  message[VIEW_ID] = viewId;
  sendToServer(message);
}// end of function sendDeleteView

// client:

// The server creates createClient and deleteClient messages.
// This approach is more robust against clients that disconnect suddenly
// or that for some other reason fail to send the right messages.

function sendUpdateClientUserName ( userName ) {
  var message = {};
  message[ACTION] = UPDATE_CLIENT;
  message[USER_NAME] = userName;
  sendToServer(message);
}// end of function sendUpdateClientUserName

function sendUpdateClientView ( viewId ) {
  var message = {};
  message[ACTION] = UPDATE_CLIENT;
  message[VIEW_ID] = viewId;
  sendToServer(message);
}// end of function sendUpdateClientView

// node:

function sendCreateNode ( nodeId, text, url, x, y, viewId ) {
  var message = {};
  message[ACTION] = CREATE_NODE;
  message[NODE_ID] = nodeId;
  message[TEXT] = text;
  message[URL] = url;
  message[X] = x;
  message[Y] = y;
  message[VIEW_ID] = viewId;
  sendToServer(message);
}// end of function sendCreateNode

function sendUpdateNodeText ( nodeId, text ) {
  var message = {};
  message[ACTION] = UPDATE_NODE;
  message[NODE_ID] = nodeId;
  message[TEXT] = text;
  sendToServer(message);
}// end of function sendUpdateNodeText

function sendUpdateNodeUrl ( nodeId, url ) {
  var message = {};
  message[ACTION] = UPDATE_NODE;
  message[NODE_ID] = nodeId;
  message[URL] = url;
  sendToServer(message);
}// end of function sendUpdateNodeUrl

function sendUpdateNodeX ( nodeId, x ) {
  var message = {};
  message[ACTION] = UPDATE_NODE;
  message[NODE_ID] = nodeId;
  message[X] = x;
  sendToServer(message);
}// end of function sendUpdateNodeX

function sendUpdateNodeY ( nodeId, y ) {
  var message = {};
  message[ACTION] = UPDATE_NODE;
  message[NODE_ID] = nodeId;
  message[Y] = y;
  sendToServer(message);
}// end of function sendUpdateNodeText

// This toggles attachment.
// If we are unattached, we attach.
// If we are attached, we deattach.
function sendUpdateNodeClient ( nodeId ) {
  var message = {};
  message[ACTION] = UPDATE_NODE;
  message[NODE_ID] = nodeId;
  sendToServer(message);
}// end of function sendUpdateNodeText

function sendDeleteNode ( nodeId ) {
  var message = {};
  message[ACTION] = DELETE_NODE;
  message[NODE_ID] = nodeId;
  sendToServer(message);
}// end of function sendDeleteNode

// edge:

function sendCreateEdge ( edgeId, subjectId, predicateId, objectId ) {
  var message = {};
  message[ACTION] = CREATE_EDGE;
  message[EDGE_ID] = edgeId;
  message[SUBJECT_ID] = subjectId;
  message[PREDICATE_ID] = predicateId;
  message[OBJECT_ID] = objectId;
  sendToServer(message);
}// end of function sendCreateEdge

function sendDeleteEdge ( edgeId ) {
  var message = {};
  message[ACTION] = DELETE_EDGE;
  message[EDGE_ID] = edgeId;
  sendToServer(message);
}// end of function sendCreateEdge

// initialization

function onPrimusError(err) {
  console.log("An error occurred in Primus: ", err);
}// end of function onPrimusError

function startPrimus () {
  primus = new Primus('127.0.0.1:8080/primus',{ });
  primus.on('open', sendSetupRequests );
  primus.on('data', onMessageFromServer);
  primus.on('error', onPrimusError);
}// end of function startPrimus

// shutdown

function endPrimus () {
  primus.end();
}// end of function endPrimus

// filter creation:

// A filter tells the server to only send this client messages that contain the specified key-value pairs.
// Besides action, the message object must have exactly one other key-value pair, filter: {.}, a filter object.
// A filter object has exactly one key-value pair:
// and: [...], or: [...], not: {.}, or any one of the key-value pairs a message can have.
// {.} must itself be a valid filter object.
// [...] must be an array of valid filter objects.
// The server tests recursively the different requirements and only sends to the client messages that meet them.

function getKeyValueFilter ( key, value ) {
  var filter = {};
  filter[key] = value;
  return filter;
}// end of getKeyValueFilter

var NOT = 'not';

function getNotFilter ( filter ) {
  var notFilter = {};
  notFilter[NOT] = filter;
  return notFilter;
}// end of getNotFilter

var AND = 'and';

function getAndFilter ( filters ) {
  var andFilter = {};
  andFilter[AND] = filters;
  return andFilter;
}// end of getAndFilter

var OR = 'or';

function getOrFilter ( filters ) {
  var orFilter = {};
  orFilter[OR] = filters;
  return orFilter;
}// end of function getOrFilter

function getViewFilter ( viewId ) {
  return getKeyValueFilter( VIEW_ID, viewId );
}// end of function getViewFilter

// The server attaches a client ID to every message it receives,
// so this filter will catch all the messages the client sent,
// even though onMessageFromServer does not always use the client ID.
function getClientFilter ( clientId ) {
  return getKeyValueFilter( CLIENT_ID, clientId );
}// end of function getClientFilter

// This will return messages altering the node itself or associating it with an edge.
function getNodeFilter ( nodeId ) {
  var filters = [];
  filters.push( getKeyValueFilter( NODE_ID, nodeId ) );
  filters.push( getKeyValueFilter( SUBJECT_ID, nodeId ) );
  filters.push( getKeyValueFilter( PREDICATE_ID, nodeId ) );
  filters.push( getKeyValueFilter( OBJECT_ID, nodeId ) );
  return getOrFilter(filters);
}// end of function getNodeFilter

function getEdgeFilter ( edgeId ) {
  return getKeyValueFilter( EDGE_ID, edgeId );
}// end of function getEdgeFilter
