// This script references variables and functions in shared.js.

// Distance functions: take in two variables and output a number representing how different they are.

// Use this distance function for strings.
// Returns the minimum number of insertions, deletions, replacements, and transpositions needed to convert string1 to string2.
// Please see https://en.wikipedia.org/wiki/Damerau%E2%80%93Levenshtein_distance.
function computeDamerauLevenshteinDistance (string1, string2) {
  // Make sure we are working with strings.
  string1 = string1.toString();
  string2 = string2.toString();
  var numRows = string1.length+1;
  var numCols = string2.length+2;
  var distances = [];
  var r = 0;
  var c = 0;
  for(r = 0; r < numRows; r++) {
    distances[r] = [];
    distances[r][0] = r;
  }
  for(c = 1; c < numCols; c++){
    distances[0][c] = c;
  }
  var string1Index = 0;
  var string2Index = 0;
  for(r = 1; r < numRows; r++) {
    string1Index = r-1;
    for(c = 1; c < numCols; c++) {
      string2Index = c-1;
      var substitutionCost = 1;
      if ( string1[string1Index] == string2[string2Index] ) {
        substitutionCost = 0;
      }
      //                          insertion,           deletion,            substitution
      distances[r][c] = Math.min( distances[r-1][c]+1, distances[r][c-1]+1, distances[r-1][c-1]+substitutionCost );
      // transposition
      if ( (string1Index > 0) && (string2Index > 0) ) {
        if ( (string1[string1Index] == string2[string2Index-1]) && (string1[string1Index-1] == string2[string2Index]) ) {
          distances[r][c] = Math.min( distances[r][c], distances[r-2][c-2] + substitutionCost );
        }// end of if the current and previous characters are transposed between string1 and string2
      }// end of if we have earlier letters to transpose
    }// end of loop through columns in row
  }// end of loop through rows
  return distances[numRows-1][numCols-1];
}// end of function computeDamerauLevenshteinDistance

// Use this distance function for numbers.
function computeAbsDifference ( num1, num2 ) {
  // Make sure we are working with numbers.
  return Math.abs( Number(num2) - Number(num1) );
}// end of function computeAbsDifference

// item class used inside ListSortedClosestFirst:
function SortableItem ( id, value, distanceFromTarget, index ) {
  this.id = id;
  this.value = value;
  this.distanceFromTarget = distanceFromTarget;
  this.index = index;
}// end of function SortableItem

// maintains a list of items with IDs mapped to values in a one-to-one relationship
// Each value is dome distance from a shared target value,
// and the items are sorted with items closer to the target at lower indeces.
function ListSortedClosestFirst ( computeDistance, target ) {
  
  var itemById = {};
  var itemByIndex = [];
  
  var target = target;
  
  var cells = [];
  
  // Store the index in itemByIndex of each item in the item.index property.
  function storeIndeces () {
    function storeIndex ( item, index ) {
      item.index = index;
    }// end of function storeIndex
    itemByIndex.forEach(storeIndex);
  }// end of function storeIndeces
  
  // Set a new target and redo the sorting.
  this.setTarget = function ( newTarget ) {
    
    target = newTarget;
    
    // Recompute the distance of each value from the target.
    function updateDistanceFromTarget ( item ) {
      item.distanceFromTarget = computeDistance( item.value, target );
    }// end of function updateDistanceFromTarget
    itemByIndex.forEach( updateDistanceFromTarget );
    
    // Then resort the whole list, which involves multiple comparissons with each distance.
    // A is closer than B. => dist(A) - dist(B) < 0 => Put A before (at lower index than) B.
    // A is the same distance as B. => dist(A) - dist(B) == 0 => Keep ordering the same. (Not all browsers follow this rule.)
    // A is further than B. => dist(A) - dist(B) > 0 => Put A after (at higher index than) B.
    function compareCloseness ( itemA, itemB ) {
      return itemA.distanceFromTarget - itemB.distanceFromTarget;
    }// end of function compareCloseness
    itemByIndex.sort( compareCloseness );
    
    // Store the updated indeces.
    storeIndeces();
    
  };// end of this.setTarget
  
  // Call this to map an ID to a value.
  // If the ID is already mapped to another value, the new value replaces the old one.
  // In either case, the item gets placed at its proper index in the list.
  this.setItem = function ( id, value ) {
    
    // Get rid of both references to the item.
    if ( itemById.hasOwnProperty(id) ) {
      itemByIndex.splice( itemById[id].index, 1 );
      delete itemById[id];
    }
    
    // Create a new one.
    var newItem = new SortableItem( id, value, computeDistance(value, target), -1 );
    
    // Store a reference at the correct ID.
    itemById[id] = newItem;
    
    // Find the correct index.
    function findIndex ( firstIndex, segmentLength ) {
      if ( segmentLength === 0 ) {
        return firstIndex;
      }
      else if ( segmentLength === 1 ) {
        if ( newItem.distanceFromTarget < itemByIndex[firstIndex].distanceFromTarget ) {
          return firstIndex;
        }
        else {
          return firstIndex+1;
        }
      }
      else {
        var subSegmentLength1 = Math.floor(segmentLength/2);
        var pivotIndex = firstIndex + subSegmentLength1;
        if ( newItem.distanceFromTarget < itemByIndex[pivotIndex].distanceFromTarget ) {
          // Compare it to cells before the pivot index.
          return findIndex( firstIndex, subSegmentLength1 );
        }
        else {
          // Compare it to cells after the pivot index.
          return findIndex( pivotIndex, segmentLength - subSegmentLength1 );
        }
      }
    }// end of function findIndex
    var newIndex = findIndex(0,itemByIndex.length);
    // Store a reference to the item at the correct index.
    itemByIndex.splice(newIndex, 0, newItem);
    
    // Store the updated indeces.
    storeIndeces();
    
  };// end of this.setItem
  
  this.unsetItem = function ( id ) {
    
    // Get rid of both references to the item.
    if ( itemById.hasOwnProperty(id) ) {
      itemByIndex.splice( itemById[id].index, 1 );
      delete itemById[id];
    }
    
    // Store the updated indeces.
    storeIndeces();
    
  };// end of this.unsetItem
  
  this.getIndex = function ( id ) {
    return itemById[id].index;
  };// end of this.getIndex
  
  // returns an array of IDs in sorted order
  this.getIdArray = function () {
    function returnItemId(item) {
      return item.id;
    }// end of function returnItemId
    return itemByIndex.map(returnItemId);
  };// end of this.getIdArray
  
}// end of function ListSortedClosestFirst

// reminders of what these arguments to table.insertRow and table.insertCell mean:

var AT_START = 0;
var AT_END = -1;

// Watchable to comparable value converters:

// In general, a function should take in two arguments,
// the old value and new value of the watched variable, 
// or null for the old value if we are just setting the cell content for the first time,
// and output a number or string suitable for use with the appropriate distance function.

// If the watched value is already of the right type, we just want to replace the old with the new.
function returnNewValue ( oldValue, newValue ) {
  return newValue;
}// end of function returnNewValue

// When we have a View, Client, Node, or Edge object,
// we want to compare its ID for similarity with other IDs.
// View, Node, and Edge already have string IDs, but Client has a number.
// For our purposes, it is most convenient to make all IDs strings,
// and use the blank string to represent null.
function returnNewIdStringOrEmptyString ( oldObject, newObject ) {
  if ( newObject === null ) {
    return "";// If the object is null, return an empty string.
  }
  else {
    return newObject.getId().toString();// Otherwise, output its ID as a string.
  }
}// end of function eturnNewIdStringOrEmptyString

// Insert to the table row a cell that changes when the watchable value changes.
function insertUpdatingCell ( row, watchable, getCellTextFromWatchedValues ) {
  var cell = row.insertCell(AT_END);
  // Initialize the cell text to the current value.
  cell.textContent = getCellTextFromWatchedValues( null, watchable.get() );
  // Set a callback so that, when the value changes, the cell text will update.
  function updateCellText ( oldValue, newValue ) {
    cell.textContent = getCellTextFromWatchedValues( null, watchable.get() );
  }// end of updateCellText
  watchable.addChangeCallback( updateCellText );
}// end of function insertUpdatingCell

function orderRows ( table, rowById, orderedList ) {
  // We remove each row and reinsert it at the end of the table,
  // so, when we are done, the rows are in the correct order.
  function reinsert ( id ) {
    var row = rowById[id];
    // appendChild removes the row DOM node from its current location
    // and places it at the end of the table.
    table.appendChild(row);
  }// end of function reinsert
  orderedList.getIdArray().forEach(reinsert);
}// end of function orderRows

function setUpUserNameField () {
  
  var userNameField = document.getElementById('userNameField');
  userNameField.oninput = function () {
    sendUpdateClientUserName( userNameField.value );
  };// end of userNameField.oninput
  
}// end of function setUpUserNameField

function setUpViewTable () {
  
  var table = document.getElementById('viewTable');
  var viewIdField = document.getElementById('viewIdField');
  var createButton = document.getElementById('createViewButton');
  var deleteButton = document.getElementById('deleteViewButton');
  
  // The view list has only one input field, the view ID field, for use in searching for items.
  var idList = new ListSortedClosestFirst(computeDamerauLevenshteinDistance, viewIdField.value);
  
  var rowById = {};
  
  function checkTargetCreatability () {
    createButton.disabled = canCreateView( viewIdField.value, new ValidationResult() ).getInvalid();
  }// end of function checkTargetCreatability
  viewMap.addCreateCallback(checkTargetCreatability);
  viewMap.addDeleteCallback(checkTargetCreatability);
  
  function checkTargetDeletability () {
    deleteButton.disabled = canDeleteView( viewIdField.value, new ValidationResult() ).getInvalid();
  }// end of function checkTargetDeletability
  viewMap.addCreateCallback(checkTargetDeletability);
  viewMap.addDeleteCallback(checkTargetDeletability);
  
  function addDeletabilityCheckOnCountChange ( id, view ) {
    view.clientCountWatch.addChangeCallback( checkTargetDeletability );
    view.nodeCountWatch.addChangeCallback( checkTargetDeletability );
    view.edgeCountWatch.addChangeCallback( checkTargetDeletability );
  }// end of function addDeletabilityCheckOnCountChange
  viewMap.addCreateCallback(addDeletabilityCheckOnCountChange);
  
  function insertViewRow ( id, view ) {
    idList.setItem( id, id );
    var row = table.insertRow( idList.getIndex(id) );
    rowById[id] = row;
    var idCell = row.insertCell(AT_END);
    idCell.textContent = id;
    // These counts have numerical values we can just pass through to the cell.
    // Conversion to string happens automatically.
    // We do not sort by any of them.
    insertUpdatingCell( row, view.clientCountWatch, returnNewValue );
    insertUpdatingCell( row, view.nodeCountWatch, returnNewValue );
    insertUpdatingCell( row, view.edgeCountWatch, returnNewValue );
  }// end of function insertViewRow
  viewMap.addCreateCallback(insertViewRow);
  
  function deleteRow ( id, view ) {
    table.deleteRow( idList.getIndex(id) );
    idList.unsetItem( id );
    delete rowById[id];
  }// end of function deleteRow
  viewMap.addDeleteCallback(deleteRow);

  viewIdField.oninput = function () {
    // Update button enabledness to reflect whether a view with the entered ID does or does not exist.
    checkTargetCreatability();
    checkTargetDeletability();
    // Reorder
    idList.setTarget( viewIdField.value );
    orderRows( table, rowById, idList );
  };// end of viewIdField.oninput
  
  createButton.onclick = function () {
    sendCreateView( viewIdField.value );
  };// end of createButton.onclick
  
  deleteButton.onclick = function () {
    sendDeleteView( viewIdField.value );
  };// end of deleteButton.onclick
  
}// end of function setUpViewTable

function setUpClientTable () {

  var table = document.getElementById('clientTable');
  var clientIdField = document.getElementById('clientIdField');
  var clientUserNameField = document.getElementById('clientUserNameField');
  var clientViewIdField = document.getElementById('clientViewIdField');
  
  // The client table has three text fields, client ID, user name, and view ID, for use in searching for items.
  var clientIdList = new ListSortedClosestFirst( computeDamerauLevenshteinDistance, clientIdField.value );
  var userNameList = new ListSortedClosestFirst( computeDamerauLevenshteinDistance, clientUserNameField.value );
  var viewIdList = new ListSortedClosestFirst( computeDamerauLevenshteinDistance, clientViewIdField.value );
  // We can only sort by one ordering at a time. Use currentSortList to reference the one we are currently using.
  var currentSortList = clientIdList;
  
  var rowById = {};
  
  // The user cannot create or delete clients.
  
  function insertClientRow ( id, client ) {
    clientIdList.setItem( id, id );
    userNameList.setItem( id, client.userNameWatch.get() );
    viewIdList.setItem(  id,  returnNewIdStringOrEmptyString( null, client.viewWatch.get() )  );
    var row = table.insertRow( currentSortList.getIndex(id) );
    rowById[id] = row;
    var idCell = row.insertCell(AT_END);
    idCell.textContent = id;
    // The user name and view ID both update and both can be the sorting column.
    insertUpdatingCell( row, client.userNameWatch, returnNewValue );
    insertUpdatingCell( row, client.viewWatch, returnNewIdStringOrEmptyString );
    function updateUserNameList ( oldName, newName ) {
      userNameList.setItem( id, newName );
      if ( currentSortList === userNameList ) {
        orderRows( table, rowById, currentSortList );
      }
    }// end of updateUserNameList
    client.userNameWatch.addChangeCallback(updateUserNameList);
    function updateViewIdList ( oldView, newView ) {
      viewIdList.setItem( id, returnNewIdStringOrEmptyString(oldView, newView) );
      if ( currentSortList === viewIdList ) {
        orderRows( table, rowById, currentSortList );
      }
    }// end of updateViewIdList
    client.viewWatch.addChangeCallback(updateViewIdList);
  }// end of function insertClientRow
  clientMap.addCreateCallback(insertClientRow);
  
  function deleteRow ( id, view ) {
    table.deleteRow( currentSortList.getIndex(id) );
    clientIdList.unsetItem( id );
    userNameList.unsetItem( id );
    viewIdList.unsetItem( id );
    delete rowById[id];
  }// end of function deleteRow
  clientMap.addDeleteCallback(deleteRow);
  
  clientIdField.oninput = function () {
    // Reorder
    clientIdList.setTarget( clientIdField.value );
    currentSortList = clientIdList;
    orderRows( table, rowById, currentSortList);
  };// end of viewIdField.oninput

  clientUserNameField.oninput = function () {
    // Reorder
    userNameList.setTarget( clientUserNameField.value );
    currentSortList = userNameList;
    orderRows( table, rowById, currentSortList );
  };// end of clientUserNameField.oninput

  clientViewIdField.oninput = function () {
    // Reorder
    viewIdList.setTarget( clientViewIdField.value );
    currentSortList = viewIdList;
    orderRows( table, rowById, currentSortList );
  };// end of clientViewIdField.oninput
  
}// end of function setUpClientTable

function setUpNodeTable () {

  var table = document.getElementById('nodeTable');
  var nodeViewIdField = document.getElementById('nodeViewIdField');
  var nodeIdField = document.getElementById('nodeIdField');
  var nodeTextField = document.getElementById('nodeTextField');
  var nodeUrlField = document.getElementById('nodeUrlField');
  var nodeXField = document.getElementById('nodeXField');
  var nodeYField = document.getElementById('nodeYField');
  var nodeClientIdField = document.getElementById('nodeClientIdField');
  var createButton = document.getElementById('createNodeButton');
  var toggleNodeLockCheckbox = document.getElementById('toggleNodeLockCheckbox');
  var deleteButton = document.getElementById('deleteNodeButton');
  
  var rowById = {};
  
  var viewIdList = new ListSortedClosestFirst( computeDamerauLevenshteinDistance, nodeViewIdField.value );
  var nodeIdList = new ListSortedClosestFirst( computeDamerauLevenshteinDistance, nodeIdField.value );
  var textList = new ListSortedClosestFirst( computeDamerauLevenshteinDistance, nodeTextField.value );
  var urlList = new ListSortedClosestFirst( computeDamerauLevenshteinDistance, nodeUrlField.value );
  var xList = new ListSortedClosestFirst( computeAbsDifference, Number(nodeXField.value) );
  var yList = new ListSortedClosestFirst( computeAbsDifference, Number(nodeYField.value) );
  var clientIdList = new ListSortedClosestFirst( computeDamerauLevenshteinDistance, nodeClientIdField.value );
  // We can only sort by one ordering at a time. Use currentSortList to reference the one we are currently using.
  var currentSortList = nodeIdList;
  
  function checkTargetCreatability () {
    var nodeId = nodeIdField.value;
    var text = nodeTextField.value;
    var url = nodeTextField.value;
    var x = nodeXField.value;
    var y = nodeYField.value;
    var viewId = nodeViewIdField.value;
    // var clientId = nodeClientIdField.value;-This is search only. The new node gets created with the local client's ID.
    var result = canCreateNode( nodeId, text, url, x, y, viewId, myClientId, new ValidationResult() );
    createButton.disabled = result.getInvalid();
    // console.log( "can create node invalid: ", result.getInvalid(), ", reasons: ", result.getReasons() );
  }// end of function checkTargetCreatability
  nodeMap.addCreateCallback(checkTargetCreatability);
  nodeMap.addDeleteCallback(checkTargetCreatability);
  
  function checkTargetLockability () {
    // The canUpdateNodeText checks whether we have locked the node for editing and whether the text is a string.
    // For this check, we do not care whether the text is valid.
    var result = canUpdateNodeClient( nodeIdField.value, myClientId, new ValidationResult() );
    toggleNodeLockCheckbox.disabled = result.getInvalid();
    // console.log( "can lock node invalid: ", result.getInvalid(), ", reasons: ", result.getReasons() );
  }// end of function checkTargetLockability
  nodeMap.addCreateCallback(checkTargetLockability);
  nodeMap.addDeleteCallback(checkTargetLockability);
  
  function addLockabilityCheckOnClientChange ( id, node ) {
    node.clientWatch.addChangeCallback( checkTargetLockability );
  }// end of function addLockabilityCheckOnClientChange
  nodeMap.addCreateCallback(addLockabilityCheckOnClientChange);
  
  function checkTargetDeletability () {
    var result = canDeleteNode( nodeIdField.value, myClientId, new ValidationResult() );
    deleteButton.disabled = result.getInvalid();
    // console.log( "can delete node invalid: ", result.getInvalid(), ", reasons: ", result.getReasons() );
  }// end of function checkTargetDeletability
  nodeMap.addCreateCallback(checkTargetDeletability);
  nodeMap.addDeleteCallback(checkTargetDeletability);
  
  function addDeletabilityCheckOnCountOrClientChange ( id, node ) {
    node.subjectEdgeCountWatch.addChangeCallback( checkTargetDeletability );
    node.predicateEdgeCountWatch.addChangeCallback( checkTargetDeletability );
    node.objectEdgeCountWatch.addChangeCallback( checkTargetDeletability );
    node.clientWatch.addChangeCallback( checkTargetDeletability );
  }// end of function addDeletabilityCheckOnCountOrClientChange
  nodeMap.addCreateCallback(addDeletabilityCheckOnCountOrClientChange);
  
  function setEditMode ( targetNode ) {
    // Set the fields to the current properties of the node the user is editing.
    nodeIdField.value = targetNode.getId();
    nodeViewIdField.value = targetNode.viewWatch.get().getId();
    nodeTextField.value = targetNode.textWatch.get();
    nodeUrlField.value = targetNode.urlWatch.get();
    nodeXField.value = targetNode.xWatch.get();
    nodeYField.value = targetNode.yWatch.get();
    nodeClientIdField.value = targetNode.clientWatch.get().getId();
    // Do not allow the user to change the node ID until it releases the lock.
    nodeIdField.disabled = true;
    toggleNodeLockCheckbox.checked = true;
    toggleNodeLockCheckbox.disabled = false;
    // These properties are not editable:
    nodeViewIdField.disabled = true;
    nodeClientIdField.disabled = true;
  }// end of function setEditMode
  
  function unsetEditMode () {
    // The fields become search fields again, so all should be editable.
    nodeIdField.disabled = false;
    toggleNodeLockCheckbox.checked = false;
    nodeViewIdField.disabled = false;
    nodeClientIdField.disabled = false;
  }// end of function unsetEditMode
  
  function insertNodeRow ( id, node ) {
    viewIdList.setItem(  id,  returnNewIdStringOrEmptyString( null, node.viewWatch.get() )  );
    nodeIdList.setItem( id, id );
    textList.setItem( id, node.textWatch.get() );
    urlList.setItem( id, node.urlWatch.get() );
    xList.setItem( id, node.xWatch.get() );
    yList.setItem( id, node.yWatch.get() );
    clientIdList.setItem(  id,  returnNewIdStringOrEmptyString( null, node.clientWatch.get() )  );
    var row = table.insertRow( currentSortList.getIndex(id) );
    rowById[id] = row;
    // The viewId, text, x, y, and client all update and can be the sorting column.
    insertUpdatingCell( row, node.viewWatch, returnNewIdStringOrEmptyString );
    var idCell = row.insertCell(AT_END);
    idCell.textContent = id;
    insertUpdatingCell( row, node.textWatch, returnNewValue );
    insertUpdatingCell( row, node.urlWatch, returnNewValue );
    insertUpdatingCell( row, node.xWatch, returnNewValue );
    insertUpdatingCell( row, node.yWatch, returnNewValue );
    insertUpdatingCell( row, node.clientWatch, returnNewIdStringOrEmptyString );
    // Set a callback to update the ordering of view IDs.
    function updateViewIdList ( oldView, newView ) {
      viewIdList.setItem( id, returnNewIdStringOrEmptyString(oldView, newView) );
      if ( currentSortList === viewIdList ) {
        orderRows( table, rowById, currentSortList );
      }
    }// end of updateViewIdList
    node.viewWatch.addChangeCallback(updateViewIdList);
    // Set a callback to update the ordering of texts.
    function updateTextList( oldText, newText ) {
      textList.setItem( id, newText );
      if ( currentSortList === textList ) {
        orderRows( table, rowById, currentSortList );
      }
    }// end of updateTextList
    node.textWatch.addChangeCallback(updateTextList);
    // Set a callback to update the ordering of URLs.
    function updateUrlList( oldUrl, newUrl ) {
      textList.setItem( id, newUrl );
      if ( currentSortList === urlList ) {
        orderRows( table, rowById, currentSortList );
      }
    }// end of updateUrlList
    node.urlWatch.addChangeCallback(updateUrlList);
    // Set a callback to update the ordering of x coordinates.
    function updateXList( oldX, newX ) {
      xList.setItem( id, newX );
      if ( currentSortList === xList ) {
        orderRows( table, rowById, currentSortList );
      }
    }// end of updateXList
    node.xWatch.addChangeCallback(updateXList);
    // Set a callback to update the ordering of y coordinates.
    function updateYList( oldY, newY ) {
      yList.setItem( id, newY );
      if ( currentSortList === yList ) {
        orderRows( table, rowById, currentSortList );
      }
    }// end of updateYList
    node.yWatch.addChangeCallback(updateYList);
    // Set a callback to update the ordering of client IDs.
    function updateClientIdList ( oldClient, newClient ) {
      clientIdList.setItem( id, returnNewIdStringOrEmptyString(oldClient, newClient) );
      if ( currentSortList === clientIdList ) {
        orderRows( table, rowById, currentSortList );
      }
    }// end of updateClientIdList
    node.clientWatch.addChangeCallback(updateClientIdList);
    // Check whether this is a node to which we are attached.
    if ( node.clientWatch.get().getId() === myClientId ) {
      setEditMode( node );
    }
    function updateNodeLock ( oldClient, newClient ) {
      if (  ( oldClient !== null ) && ( oldClient.getId() === myClientId )  ) {
        unsetEditMode();
      }
      if (  ( newClient !== null ) && ( newClient.getId() === myClientId )  ) {
        setEditMode( node );
      }
    }// end of function updateNodeLock
    node.clientWatch.addChangeCallback(updateNodeLock);
  }// end of function insertNodeRow
  nodeMap.addCreateCallback(insertNodeRow);
  
  function deleteRow ( id, node ) {
    if (  ( node.clientWatch.get() !== null ) && ( node.clientWatch.get().getId() === myClientId )  ) {
      unsetEditMode();
    }
    table.deleteRow( currentSortList.getIndex(id) );
    viewIdList.unsetItem( id );
    nodeIdList.unsetItem( id );
    textList.unsetItem( id );
    urlList.unsetItem( id );
    xList.unsetItem( id );
    yList.unsetItem( id );
    clientIdList.unsetItem( id );
    delete rowById[id];
  }// end of function deleteRow
  nodeMap.addDeleteCallback(deleteRow);

  nodeViewIdField.oninput = function () {
    // To create a node, we must have an existing view to which to add it and a non-existant node ID.
    checkTargetCreatability();
    // Reorder
    viewIdList.setTarget( nodeViewIdField.value );
    currentSortList = viewIdList;
    orderRows( table, rowById, currentSortList );
  };// end of nodeViewIdField.oninput

  nodeIdField.oninput = function () {
    // Update button enabledness to reflect whether a node with the entered ID does or does not exist.
    // To create a node, we must have an existing view to which to add it and a non-existant node ID.
    checkTargetCreatability();
    // Lockability and deletability only require an existing node ID.
    checkTargetLockability();
    checkTargetDeletability();
    // Reorder
    nodeIdList.setTarget( nodeIdField.value );
    currentSortList = nodeIdList;
    orderRows( table, rowById, currentSortList );
  };// end of nodeIdField.oninput
  
  // When the user has locked node for editing, the text, x, and y fields send messages to update the value of the locked node.
  // Otherwise, they function as search fields.
  
  nodeTextField.oninput = function () {
    if ( toggleNodeLockCheckbox.checked ) {
      sendUpdateNodeText( nodeIdField.value, nodeTextField.value );
    }
    // Reorder
    textList.setTarget( nodeTextField.value );
    currentSortList = textList;
    orderRows( table, rowById, currentSortList );
  };// end of nodeTextField.oninput
  
  nodeUrlField.oninput = function () {
    if ( toggleNodeLockCheckbox.checked ) {
      sendUpdateNodeUrl( nodeIdField.value, nodeUrlField.value );
    }
    // Reorder
    urlList.setTarget( nodeUrlField.value );
    currentSortList = urlList;
    orderRows( table, rowById, currentSortList );
  };// end of nodeUrlField.oninput

  nodeXField.oninput = function () {
    if ( toggleNodeLockCheckbox.checked ) {
      sendUpdateNodeX( nodeIdField.value, nodeXField.value );
    }
    else {
      // Reorder
      xList.setTarget( nodeXField.value );
      currentSortList = xList;
      orderRows( table, rowById, currentSortList );
    }
  };// end of nodeXField.oninput

  nodeYField.oninput = function () {
    if ( toggleNodeLockCheckbox.checked ) {
      sendUpdateNodeY( nodeIdField.value, nodeYField.value );
    }
    else {
      // Reorder
      yList.setTarget( nodeYField.value );
      currentSortList = yList;
      orderRows( table, rowById, currentSortList );
    }
  };// end of nodeYField.oninput
  
  nodeClientIdField.oninput = function () {
    // Reorder
    clientIdList.setTarget( nodeClientIdField.value );
    currentSortList = clientIdList;
    orderRows( table, rowById, currentSortList );
  };// end of nodeClientIdField.oninput
  
  createButton.onclick = function () {
    var nodeId = nodeIdField.value;
    var text = nodeTextField.value;
    var url = nodeUrlField.value;
    var x = nodeXField.value;
    var y = nodeYField.value;
    var viewId = nodeViewIdField.value;
    sendCreateNode( nodeId, text, url, x, y, viewId );
  };// end of createButton.onclick
  
  toggleNodeLockCheckbox.onchange = function () {
    sendUpdateNodeClient( nodeIdField.value );
  };// end of toggleNodeLockCheckbox.onchange
  
  deleteButton.onclick = function () {
    sendDeleteNode( nodeIdField.value );
  };// end of deleteButton.onclick
  
}// end of function setUpNodeTable

function setUpEdgeTable () {
  
  var table = document.getElementById('edgeTable');
  var edgeViewIdField = document.getElementById('edgeViewIdField');
  var edgeIdField = document.getElementById('edgeIdField');
  var edgeSubjectIdField = document.getElementById('edgeSubjectIdField');
  var edgeSubjectLockCheckbox = document.getElementById('edgeSubjectLockCheckbox');
  var edgePredicateIdField = document.getElementById('edgePredicateIdField');
  var edgePredicateLockCheckbox = document.getElementById('edgePredicateLockCheckbox');
  var edgeObjectIdField = document.getElementById('edgeObjectIdField');
  var edgeObjectLockCheckbox = document.getElementById('edgeObjectLockCheckbox');
  var createButton = document.getElementById('createEdgeButton');
  var deleteButton = document.getElementById('deleteEdgeButton');
  
  var viewIdList = new ListSortedClosestFirst( computeDamerauLevenshteinDistance, edgeViewIdField.value );
  var edgeIdList = new ListSortedClosestFirst( computeDamerauLevenshteinDistance, edgeIdField.value );
  var subjectIdList = new ListSortedClosestFirst( computeDamerauLevenshteinDistance, edgeSubjectIdField.value );
  var predicateIdList = new ListSortedClosestFirst( computeDamerauLevenshteinDistance, edgePredicateIdField.value );
  var objectIdList = new ListSortedClosestFirst( computeDamerauLevenshteinDistance, edgeObjectIdField.value );
  // We can only sort by one ordering at a time. Use currentSortList to reference the one we are currently using.
  var currentSortList = edgeIdList;
  
  var rowById = {};
  
  // Recall that, to create an edge across three nodes, all nodes must exist, and we must have the lock on all three.
  
  // Whether we can create or delete an edge depends on
  // whether it itself already exists,
  // whether all three component nodes exist, and
  // whether we have locked all three component nodes.
  function checkTargetCreatability () {
    // var viewId = edgeViewIdField.value;--not set directly by the user, since it must match those of the nodes
    var edgeId = edgeIdField.value;
    var subjectId = edgeSubjectIdField.value;
    var predicateId = edgePredicateIdField.value;
    var objectId = edgeObjectIdField.value;
    var result = canCreateEdge ( edgeId, subjectId, predicateId, objectId, myClientId, new ValidationResult() );
    createButton.disabled = result.getInvalid();
    // console.log( "can can create edge invalid: ", result.getInvalid(), ", reasons: ", result.getReasons() );
  }// end of function checkTargetCreatability
  // Both node and edge creation and deletion affect what edges we can create or delete.
  edgeMap.addCreateCallback(checkTargetCreatability);
  edgeMap.addDeleteCallback(checkTargetCreatability);
  nodeMap.addCreateCallback(checkTargetCreatability);
  nodeMap.addDeleteCallback(checkTargetCreatability);
  
  // Lock events affect which nodes are currently locked.
  function addCreatabilityCheckOnClientChange ( id, node ) {
    node.clientWatch.addChangeCallback( checkTargetCreatability );
  }// end of function addLockabilityCheckOnClientChange
  nodeMap.addCreateCallback(addCreatabilityCheckOnClientChange);
  
  // Add callbacks to check whether we can lock or unlock each node.
  
  // Creation and deletion events affect which of the nodes exist.
  function checkTargetLockability () {
    edgeSubjectLockCheckbox.disabled = canUpdateNodeClient( edgeSubjectIdField.value, myClientId, new ValidationResult() ).getInvalid();
    edgePredicateLockCheckbox.disabled = canUpdateNodeClient( edgePredicateIdField.value, myClientId, new ValidationResult() ).getInvalid();
    edgeObjectLockCheckbox.disabled = canUpdateNodeClient( edgeObjectIdField.value, myClientId, new ValidationResult() ).getInvalid();
  }// end of function checkTargetLockability
  nodeMap.addCreateCallback(checkTargetLockability);
  nodeMap.addDeleteCallback(checkTargetLockability);
  
  // Lock events affect which ones are already locked.
  function addLockabilityCheckOnClientChange ( id, node ) {
    node.clientWatch.addChangeCallback( checkTargetLockability );
  }// end of function addLockabilityCheckOnClientChange
  nodeMap.addCreateCallback(addLockabilityCheckOnClientChange);
  
  // Deletability likewise depends on whether we have a lock on all three nodes.
  function checkTargetDeletability () {
    var result = canDeleteEdge( edgeIdField.value, myClientId, new ValidationResult() );
    deleteButton.disabled = result.getInvalid();
    // console.log( "can can delete edge invalid: ", result.getInvalid(), ", reasons: ", result.getReasons() );
  }// end of function checkTargetDeletability
  // We do not need to check whether individual nodes exist.
  // If the edge exists, all three nodes exist, so deletability depends on whether we have locks on all three.
  // If not, then the edge is not deletable.
  edgeMap.addCreateCallback(checkTargetDeletability);
  edgeMap.addDeleteCallback(checkTargetDeletability);
  
  // Lock events affect which ones are already locked.
  function addDeletabilityCheckOnClientChange ( id, node ) {
    node.clientWatch.addChangeCallback( checkTargetDeletability );
  }// end of function addLockabilityCheckOnClientChange
  nodeMap.addCreateCallback(addDeletabilityCheckOnClientChange);
  
  function checkEdgeNodeMode ( nodeIdField, nodeLockCheckbox, node ) {
    if ( nodeIdField.value == node.getId() ) {
      if (  ( node.clientWatch.get() !== null ) && ( node.clientWatch.get().getId() === myClientId )  ) {
        nodeIdField.disabled = true;
        nodeLockCheckbox.disabled = false;
        nodeLockCheckbox.checked = true;
      }
      else {
        nodeIdField.disabled = false;
        nodeLockCheckbox.checked = false;
      }
    }
  }// end of function checkEdgeNodeMode
  
  function updateEdgeNodeInterfaces ( id, node ) {
    checkEdgeNodeMode( edgeSubjectIdField, edgeSubjectLockCheckbox, node );
    checkEdgeNodeMode( edgePredicateIdField, edgePredicateLockCheckbox, node );
    checkEdgeNodeMode( edgeObjectIdField, edgeObjectLockCheckbox, node );
  }// end of function updateEdgeNodeInterfaces
  nodeMap.addCreateCallback(updateEdgeNodeInterfaces);
  nodeMap.addDeleteCallback(updateEdgeNodeInterfaces);
  
  function addUpdateEdgeNodeInterfacesOnClientChange ( id, node ) {
    function updateEdgeNodeInterfacesOnClientChange ( oldClient, newClient ) {
      updateEdgeNodeInterfaces( id, node );
    }// end of function updateEdgeNodeInterfacesOnClientChange
    node.clientWatch.addChangeCallback( updateEdgeNodeInterfacesOnClientChange );
  }// end of addUpdateEdgeNodeInterfacesOnClientChange
  nodeMap.addCreateCallback(addUpdateEdgeNodeInterfacesOnClientChange);
  
  
  function insertEdgeRow ( id, edge ) {
    viewIdList.setItem(  id,  returnNewIdStringOrEmptyString( null, edge.viewWatch.get() )  );
    edgeIdList.setItem( id, id );
    subjectIdList.setItem(  id,  returnNewIdStringOrEmptyString( null, edge.subjectWatch.get() )  );
    predicateIdList.setItem(  id,  returnNewIdStringOrEmptyString( null, edge.predicateWatch.get() )  );
    objectIdList.setItem(  id,  returnNewIdStringOrEmptyString( null, edge.objectWatch.get() )  );
    var row = table.insertRow( currentSortList.getIndex(id) );
    rowById[id] = row;
    // The viewId, text, x, y, and client all update and can be the sorting column.
    insertUpdatingCell( row, edge.viewWatch, returnNewIdStringOrEmptyString );
    var idCell = row.insertCell(AT_END);
    idCell.textContent = id;
    insertUpdatingCell( row, edge.subjectWatch, returnNewIdStringOrEmptyString );
    insertUpdatingCell( row, edge.predicateWatch, returnNewIdStringOrEmptyString );
    insertUpdatingCell( row, edge.objectWatch, returnNewIdStringOrEmptyString );
    // Set a callback to update the ordering of view IDs.
    function updateViewIdList ( oldView, newView ) {
      viewIdList.setItem( id, returnNewIdStringOrEmptyString(oldView, newView) );
      if ( currentSortList === viewIdList ) {
        orderRows( table, rowById, currentSortList );
      }
    }// end of updateViewIdList
    edge.viewWatch.addChangeCallback(updateViewIdList);
    // Set a callback to update the ordering of subject IDs.
    function updateSubjectIdList ( oldNode, newNode ) {
      subjectIdList.setItem( id, returnNewIdStringOrEmptyString(oldNode, newNode) );
      if ( currentSortList === subjectIdList ) {
        orderRows( table, rowById, currentSortList );
      }
    }// end of updateSubjectIdList
    edge.subjectWatch.addChangeCallback(updateSubjectIdList);
    // Set a callback to update the ordering of predicate IDs.
    function updatePredicateIdList ( oldNode, newNode ) {
      predicateIdList.setItem( id, returnNewIdStringOrEmptyString(oldNode, newNode) );
      if ( currentSortList === predicateIdList ) {
        orderRows( table, rowById, currentSortList );
      }
    }// end of updatePredicateIdList
    edge.predicateWatch.addChangeCallback(updatePredicateIdList);
    // Set a callback to update the ordering of object IDs.
    function updateObjectIdList ( oldNode, newNode ) {
      objectIdList.setItem( id, returnNewIdStringOrEmptyString(oldNode, newNode) );
      if ( currentSortList === objectIdList ) {
        orderRows( table, rowById, currentSortList );
      }
    }// end of updateObjectIdList
    edge.objectWatch.addChangeCallback(updateObjectIdList);
  }// end of function insertNodeRow
  edgeMap.addCreateCallback(insertEdgeRow);
  
  function deleteRow ( id, view ) {
    table.deleteRow( currentSortList.getIndex(id) );
    viewIdList.unsetItem( id );
    edgeIdList.unsetItem( id );
    subjectIdList.unsetItem( id );
    predicateIdList.unsetItem( id );
    objectIdList.unsetItem( id );
    delete rowById[id];
  }// end of function deleteRow
  edgeMap.addDeleteCallback(deleteRow);

  edgeViewIdField.oninput = function () {
    // Reorder
    checkTargetLockability();
    checkTargetCreatability();
    checkTargetDeletability();
    viewIdList.setTarget( edgeViewIdField.value );
    currentSortList = edgeIdList;
    orderRows( table, rowById, currentSortList );
  };// end of edgeViewIdField.oninput

  edgeIdField.oninput = function () {
    // Update button enabledness to reflect whether an edge with the entered ID does or does not exist.
    if (  edgeMap.has( edgeIdField.value )  ) {
      var edge = edgeMap.get( edgeIdField.value );
      edgeSubjectIdField.value = edge.subjectWatch.get().getId();
      checkEdgeNodeMode( edgeSubjectIdField, edgeSubjectLockCheckbox, edge.subjectWatch.get() );
      edgePredicateIdField.value = edge.predicateWatch.get().getId();
      checkEdgeNodeMode( edgePredicateIdField, edgePredicateLockCheckbox, edge.predicateWatch.get() );
      edgeObjectIdField.value = edge.objectWatch.get().getId();
      checkEdgeNodeMode( edgeObjectIdField, edgeObjectLockCheckbox, edge.objectWatch.get() );
    }
    checkTargetLockability();
    checkTargetCreatability();
    checkTargetDeletability();
    // Reorder
    edgeIdList.setTarget( edgeIdField.value );
    currentSortList = edgeIdList;
    orderRows( table, rowById, currentSortList );
  };// end of edgeIdField.oninput

  edgeSubjectIdField.oninput = function () {
    checkTargetLockability();
    checkTargetCreatability();
    checkTargetDeletability();
    if ( nodeMap.has(edgeSubjectIdField.value) ) {
      checkEdgeNodeMode( edgeSubjectIdField, edgeSubjectLockCheckbox, nodeMap.get(edgeSubjectIdField.value) );
    }
    // Reorder
    subjectIdList.setTarget( edgeSubjectIdField.value );
    currentSortList = subjectIdList;
    orderRows( table, rowById, currentSortList );
  };// end of edgeSubjectIdField.oninput

  edgePredicateIdField.oninput = function () {
    // Reorder
    checkTargetLockability();
    checkTargetCreatability();
    checkTargetDeletability();
    if ( nodeMap.has(edgePredicateIdField.value) ) {
      checkEdgeNodeMode( edgePredicateIdField, edgePredicateLockCheckbox, nodeMap.get(edgePredicateIdField.value) );
    }
    predicateIdList.setTarget( edgePredicateIdField.value );
    currentSortList = predicateIdList;
    orderRows( table, rowById, currentSortList );
  };// end of edgePredicateIdField.oninput

  edgeObjectIdField.oninput = function () {
    // Reorder
    checkTargetLockability();
    checkTargetCreatability();
    checkTargetDeletability();
    if ( nodeMap.has(edgeObjectIdField.value) ) {
      checkEdgeNodeMode( edgeObjectIdField, edgeObjectLockCheckbox, nodeMap.get(edgeObjectIdField.value) );
    }
    objectIdList.setTarget( edgeObjectIdField.value );
    currentSortList = objectIdList;
    orderRows( table, rowById, currentSortList );
  };// end of edgeObjectIdField.oninput
  
  edgeSubjectLockCheckbox.onchange = function () {
    sendUpdateNodeClient( edgeSubjectIdField.value );
    // Do not allow the user to change the node ID until it releases the lock.
    edgeSubjectIdField.disabled = edgeSubjectLockCheckbox.checked;
  };// end of edgeSubjectLockCheckbox.onchange
  
  edgePredicateLockCheckbox.onchange = function () {
    sendUpdateNodeClient( edgePredicateIdField.value );
    // Do not allow the user to change the node ID until it releases the lock.
    edgePredicateIdField.disabled = edgePredicateLockCheckbox.checked;
  };// end of edgePredicateLockCheckbox.onchange
  
  edgeObjectLockCheckbox.onchange = function () {
    sendUpdateNodeClient( edgeObjectIdField.value );
    // Do not allow the user to change the node ID until it releases the lock.
    edgeObjectIdField.disabled = edgeObjectLockCheckbox.checked;
  };// end of edgeObjectLockCheckbox.onchange
  
  createButton.onclick = function () {
    var edgeId = edgeIdField.value;
    var subjectId = edgeSubjectIdField.value;
    var predicateId = edgePredicateIdField.value;
    var objectId = edgeObjectIdField.value;
    sendCreateEdge( edgeId, subjectId, predicateId, objectId );
  };// end of createButton.onclick
  
  deleteButton.onclick = function () {
    sendDeleteEdge( edgeIdField.value );
  };// end of deleteButton.onclick
  
}// end of function setUpEdgeTable

function onLoadTables () {
  
  setUpUserNameField();
  setUpViewTable();
  setUpClientTable();
  setUpNodeTable();
  setUpEdgeTable();
  startPrimus();
  
}// end of onLoadTables

function onUnloadTables () {
  
  endPrimus();
  
}// end of onUnloadTables
