'use strict';

/* This Node.js module wraps a server, preferably one created using the Primus package. */
/* It maintains a set of message queues for outgoing messages and passes incoming messages to a callback. */

// Note: In primus, spark IDs are integers.

var inited = false;

// Set these by passing values in to init:
var primusServer = null;
var onNewClient = null;
var onIncomingMessage = null;
var onClosedClient = null;

// will store an object where spark IDs map to arrays of messages:
var outgoingMessagesByRecipientId = null;

// onConnection, onError, and onDisconnection are event handlers for the events,
// 'connection', 'error', and 'disconnection' from the primus server instance, 
// respectively.

function onConnection (spark) {
  console.log('connection opened: ', spark);
  
  var sparkId = spark.id;
  // Make an empty array for outgoing messages to this spark:
  outgoingMessagesByRecipientId[sparkId] = [];
  
  // When we get an incoming message, send it along with the ID of the sender:
  function onData ( message ) {
    if(onIncomingMessage !== null) {
      onIncomingMessage(sparkId, message);
    }
  }// end of function onData
  spark.on('data', onData);
  
  // When the connection ends permanently, remove the message queue.
  function onEnd () {
    // We call onClosedClient after removing the message queue.
    // This keeps any calls to dispatcher from trying to dispatch messages to the closed spark.
    delete outgoingMessagesByRecipientId[sparkId];
    onClosedClient(sparkId);
  }// end of function onEnd
  spark.on('end', onEnd);
  
  // We call onNewClient after setting up the message queue and everything else.
  // This ensures any calls to dispatcher methods in onNewClient can send messages to the new spark.
  onNewClient(sparkId);
  
}// end of function onConnection

function onError (err) {
  //Report any errors.
  console.error('error from primus: ', err);
}// end of function onError

function onDisconnection (spark) {
  // We currently use the 'end' event associated with a spark to close it,
  // instead of the disconnection event.
  // This is because the spark may temporarily disconnect and reconnect.
  // Even so, we can still report the disconnection.
  console.log('connection closed: ', spark);
}// end of function onDisconnection

// Call before trying to use dispatcher.
// newPrimusServer: a server returned by primus.createServer(...) or new Primus(...)
// None of the primus options matter for the code in this module.
// newOnNewClient, newOnIncomingMessage, newOnClosedClient should be function callbacks.
// We do not expect any return values.
// newOnNewClient and newOnClosedClient should each take a single integer ID as the only argument.
// newOnIncomingMessage should take an integer ID and one of
// whatever kind of object the server expects to get from the client.
exports.init = function (newPrimusServer, newOnNewClient, newOnIncomingMessage, newOnClosedClient) {
  if (inited) {
    throw( new Error("dispatcher.init called while dispatcher is already initiated.") );
  }
  primusServer      = newPrimusServer;
  onNewClient       = newOnNewClient;
  onIncomingMessage = newOnIncomingMessage;
  onClosedClient    = newOnClosedClient;
  outgoingMessagesByRecipientId = {};
  primusServer.on('connection', onConnection);
  primusServer.on('error', onError);
  primusServer.on('disconnection', onDisconnection);
  inited = true;
};// end of exports.init

// queues the message, which should be a serializable object
exports.queueOutgoing = function ( recipientId, message ) {
  if (!inited) {
    throw( new Error("dispatcher.queueOutgoing called while dispatcher is not initiated.") );
  }
  if ( !outgoingMessagesByRecipientId.hasOwnProperty(recipientId) ) {
    throw( new Error("queueOutgoing called with non-existent recipientId: "+recipientId) );
  }
  outgoingMessagesByRecipientId[recipientId].push(message);
};// end of exports.queueOutgoing

// removes a queued message that has already been queued
exports.unqueueOutgoing = function ( recipientId, message ) {
  if (!inited) {
    throw( new Error("dispatcher.unqueueOutgoing called while dispatcher is not initiated.") );
  }
  if ( !outgoingMessagesByRecipientId.hasOwnProperty(recipientId) ) {
    throw( new Error("unqueueOutgoing called with non-existent recipientId: "+recipientId) );
  }
  var messageIndex = outgoingMessagesByRecipientId[recipientId].indexOf(message);
  if ( messageIndex == -1 ) {
    throw( new Error("unqueueOutgoing called with non-existent message for recipientId "+recipientId) );
  }
  outgoingMessagesByRecipientId[recipientId].splice(messageIndex, 1);
};// end of exports.unqueueOutgoing

function dispatchOutgoingToSpark ( recipientSpark ) {
  function send ( message ) {
    recipientSpark.write(message);
  }// end of function send
  var recipientId = recipientSpark.id;
  outgoingMessagesByRecipientId[recipientId].forEach(send);
  outgoingMessagesByRecipientId[recipientId] = [];// Reset the queue to empty.
}// end of function dispatchOutgoingToSpark

// Sends out all queued messages and empties the queues.
exports.dispatchOutgoing = function() {
  if (!inited) {
    throw( new Error("dispatcher.dispatchOutgoing called while dispatcher is not initiated.") );
  }
  primusServer.forEach(dispatchOutgoingToSpark);
};// end of exports.dispatchOutgoing

// Call this to clean up the dispatcher in an orderly manner.
exports.end = function () {
  if (!inited) {
    throw( new Error("dispatcher.end called while dispatcher is not initiated.") );
  }
  inited = false;
  primusServer.removeListener('disconnection', onDisconnection);
  primusServer.removeListener('error', onError);
  primusServer.removeListener('connection', onConnection);
  onClosedClient    = null;
  onIncomingMessage = null;
  onNewClient       = null;
  primusServer      = null;
};// end of exports.end